import numpy as np
import logging

log = logging.getLogger("fwl."+__name__)

def find_intersection(curve1, curve2, tolerance=1e-1):
    # Starting from the TOA, find the last index where the absolute difference is within the tolerance
    intersection_indices = np.where(np.abs(curve1 - curve2) < tolerance)[-1]
    if len(intersection_indices) > 0:
        # Find the first occurrence of intersection
        first_intersection_index = intersection_indices[-1]
        first_intersection_value = curve1[first_intersection_index]
        
        return first_intersection_index, first_intersection_value
    else:
        log.error("No LCL found")
        return None


import matplotlib as mpl
mpl.use('Agg')

import matplotlib.pyplot as plt
import numpy as np
from janus.utils.atmosphere_column import atmos
import janus.utils.phys as phys

def plot_emission(atm:atmos, filename:str='output/toa_emission.pdf', 
                  level_idx:int=0, planck_surface:bool=True, 
                  incoming_solar:bool=True, show_bands:bool=False):

    # Configure
    planck_samps = 1000
    lw = 0.5
    xmin = 0.1 # nm
    xmax = 3e5   # nm

    # Get data
    x = np.zeros(atm.nbands)
    y = np.zeros(atm.nbands)

    for i in range(atm.nbands):

        # Wavelength
        x[i] += atm.band_centres[i]

        # Flux 
        y[i] += atm.LW_spectral_flux_up[i,level_idx]  # W m-2
        y[i] += atm.SW_spectral_flux_up[i,level_idx]  # W m-2
        y[i] /= atm.band_widths[i]  # W m-2 nm-1
        y[i] *= 1e3  # erg s-1 cm-2 nm-1


    # Make plot 
    fig,ax = plt.subplots(1,1, figsize=(8,4))

    # Plot incoming stellar spectrum
    if incoming_solar:
        ysolar = atm.SW_spectral_flux_down.T[0] # W m-2 
        ysolar /= atm.band_widths 
        ysolar *= 1e3 
        ax.plot(x, ysolar, color="green", lw=lw, label="TOA stellar flux")

    # Plot band edges
    if show_bands:
        for i in range(atm.nbands):
            c = 'pink'
            l = ""
            if i == 0:
                l = "Band edges"
            ax.axvline(x=atm.band_edges[i]  , color=c, lw=0.2, label=l)
            ax.axvline(x=atm.band_edges[i+1], color=c, lw=0.2)
    
    # Plot surface emission
    if planck_surface:
        planck_tmp = atm.ts
        xp = np.logspace(np.log10(np.amin(x))  ,  np.log10(np.amax(x))  ,  planck_samps)
        yp = np.zeros(np.shape(xp))
        for i in range(planck_samps):
            wl = xp[i] * 1.0e-9 # metres

            # Calculate planck function value [W m-2 sr-1 m-1]
            # http://spiff.rit.edu/classes/phys317/lectures/planck.html
            yp[i] = 2.0 * phys.h * phys.c**2.0 / wl**5.0   *   1.0 / ( np.exp(phys.h * phys.c / (wl * phys.k * planck_tmp)) - 1.0)

            # Integrate solid angle (hemisphere), scale by albedo, convert units
            yp[i] = yp[i] * np.pi * 1.0e-9 # [W m-2 nm-1]
            yp[i] = yp[i] * (1.0 - atm.albedo_s)
            yp[i] = yp[i] * 1000.0 # [erg s-1 cm-2 nm-1]
        ax.plot(xp,yp, color="dodgerblue", lw=lw, label="Surface")
    
    # Plot spectrum
    ax.plot(x,y, color='k', lw=lw, label="Emission ($i=%d$)"%level_idx)

    # Adjust
    ax.set_xlabel("Wavelength [nm]")
    ax.set_xlim(max(xmin, np.amin(x)), min(xmax, np.amax(x)))
    ax.set_xscale("log")

    ax.set_ylabel("Flux density [erg s$^{-1}$ cm$^{-2}$ nm$^{-1}$]")
    ax.set_yscale("log")

    ax.legend(loc="lower center")

    wlwn = lambda x: 1.0e7/(x-1.0e-20)  # wavelength <-> wavenumber, with small offset to prevent divergence
    secax = ax.secondary_xaxis('top', functions=(wlwn, wlwn))
    secax.set_xlabel('Wavenumber [cm$^{-1}$]')

    # Save fig 
    fig.savefig(filename, bbox_inches='tight', dpi=190)
    plt.close()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 23 13:05:00 2023

@authors: 
Tim Lichtenberg (TL)    
Ryan Boukrouche (RB)
Harrison Nicholls (HN)
"""

import scipy.optimize as optimise

import logging 
log = logging.getLogger("fwl."+__name__)

from janus.modules.compute_moist_adiabat import compute_moist_adiabat
from janus.modules.dry_adiabat_timestep import compute_dry_adiabat
from janus.utils.atmosphere_column import atmos

def RadConvEqm(dirs, time, atm, standalone:bool, cp_dry:bool, trppD:bool, rscatter:bool,
               pure_steam_adj=False, surf_dt=False, cp_surf=1e5, mix_coeff_atmos=1e6, mix_coeff_surf=1e6):
    """Sets the atmosphere to a temperature profile using the general adiabat. 
    
    Optionally does radiative time-stepping, but this is deprecated.

    Parameters
    ----------
        dirs : dict
            Named directories
        time : dict
            Dictionary of time values, including stellar age and evolution of planet
        atm : atmos
            Atmosphere object from atmosphere_column.py
        standalone : bool
            Running JANUS as standalone code?
        cp_dry : bool
            Compute dry adiabat case
        trppD : bool 
            Calculate tropopause dynamically?
        rscatter : bool
            Include rayleigh scattering?
        pure_steam_adj : bool
            Use pure steam adjustment?
        surf_dt : float
            Timestep to use for T_surf timestepping cases
        cp_surf : float
            Surface heat capacity in T_surf timestepping cases
        mix_coeff_atmos : float
            Mixing coefficient (atmosphere) for T_surf timestepping cases?
        mix_coeff_surf : float
            Mixing coefficient (surface) for T_surf timestepping cases?
            
    """

    ### Moist/general adiabat

    atm_moist = compute_moist_adiabat(atm, dirs, standalone, trppD, rscatter)

    ### Dry adiabat
    if cp_dry == True:

        # Compute dry adiabat  w/ timestepping
        atm_dry   = compute_dry_adiabat(atm, dirs, standalone, rscatter, pure_steam_adj, surf_dt, cp_surf, mix_coeff_atmos, mix_coeff_surf)

        if standalone == True:
            log.info("Net, OLR => moist: %.3f, %.3f W/m^2" % ( atm_moist.net_flux[0],atm_moist.LW_flux_up[0]))
            log.info("^^       =>   dry: %.3f, %.3f W/m^2" % ( atm_dry.net_flux[0],  atm_dry.LW_flux_up[0]))
            log.info("")
    else: 
        atm_dry = {}
    
    # Plot
    # if standalone == True:
    #     #plot_flux_balance(atm_dry, atm_moist, cp_dry, time, dirs)
    #     # Save to disk
    #     with open(dirs["output"]+"/"+str(int(time["planet"]))+"_atm.pkl", "wb") as atm_file: 
    #         pkl.dump(atm_moist, atm_file, protocol=pkl.HIGHEST_PROTOCOL)

    return atm_dry, atm_moist


def MCPA(dirs, atm, standalone:bool, trppD:bool, rscatter:bool):
    """Calculates the temperature profile using the multiple-condensible pseudoadiabat.

    Prescribes a stratosphere, and also calculates fluxes.

    Parameters
    ----------
        dirs : dict
            Named directories
        atm : atmos
            Atmosphere object from atmosphere_column.py
        standalone : bool
            Running JANUS as standalone code?
        trppD : bool 
            Calculate tropopause dynamically?
        rscatter : bool
            Include rayleigh scattering?
            
    """

    ### Moist/general adiabat
    return compute_moist_adiabat(atm, dirs, standalone, trppD, rscatter)

def MCPA_CBL(dirs, atm_inp, trppD:bool, rscatter:bool, atm_bc:int=0, T_surf_guess:float=-1, T_surf_max:float=-1, method:int=0, atol:float=1.0e-3):
    """Calculates the temperature profile using the multiple-condensible pseudoadiabat and steps T_surf to conserve energy.

    Prescribes a stratosphere, and also calculates fluxes. Finds T_surf and fluxes such that the conductive BL conduction
    equation is satisfied (Fourier's law).

    Parameters
    ----------
        dirs : dict
            Named directories
        atm_inp : atmos
            Atmosphere object from atmosphere_column.py
        trppD : bool 
            Calculate tropopause dynamically?
        rscatter : bool
            Include rayleigh scattering?
        
        atm_bc : int
            Where to measure boundary condition flux (0: TOA, 1: Surface).
        T_surf_guess : float
            Surface temperature guess (-1 to disable)
        T_surf_max : float
            Surface temperature ceiling (-1 to disable)
        method : int
            Root finding method (0: secant, 1: brentq)
        atol : float
            Tolerance for flux conservation [W m-2]
    """

    # Store constants
    #    Passed into atmos() constructor ...
    trppT = atm_inp.trppT
    band_edges = atm_inp.band_edges
    minT = atm_inp.minT; maxT = atm_inp.maxT
    Psurf = atm_inp.ps; ptop = atm_inp.ptop
    pl_r = atm_inp.planet_radius; pl_m = atm_inp.planet_mass
    vol_list = atm_inp.vol_list
    nlev_save = atm_inp.nlev_save
    re=atm_inp.effective_radius
    lwm=atm_inp.liquid_water_fraction
    clfr=atm_inp.cloud_fraction
    do_cloud=atm_inp.do_cloud
    alpha_cloud=atm_inp.alpha_cloud

    #    Passed later ...
    attrs = {}
    for a in ["instellation", "zenith_angle", "albedo_pl", 
                "inst_sf", "skin_k", "skin_d", "tmp_magma", "albedo_s",
                "planet_mass", "planet_radius"]:
        attrs[a] = getattr(atm_inp,a)

    # Calculate conductive flux for a given atmos object 'a'
    def skin(a):
        return a.skin_k / a.skin_d * (a.tmp_magma - a.ts)
    
    # Initialise a new atmos object
    def ini_atm(Ts):
        _atm = atmos(Ts, Psurf, ptop, 
                   pl_r, pl_m, band_edges,
                   vol_mixing=vol_list, trppT=trppT, 
                   minT=minT, maxT=maxT, 
                   req_levels=nlev_save, alpha_cloud=alpha_cloud,
                   re=re, lwm=lwm, clfr=clfr, do_cloud=do_cloud)
        for a in attrs.keys():
            setattr(_atm,a,attrs[a])
        return _atm
    
    # We want to optimise this function (returns residual of F_atm and F_skn, given T_surf)
    def func(x):

        log.info("Evaluating at T_surf = %.1f K" % x)
        atm_tmp = compute_moist_adiabat(ini_atm(x), dirs, False, trppD, rscatter)

        if atm_bc == 0:
            F_atm = atm_tmp.net_flux[0]  
        else:
            F_atm = atm_tmp.net_flux[-1]  

        F_skn = skin(atm_tmp)
        log.info("    F_atm = %+.2e W m-2      F_skn = %+.2e W m-2" % (F_atm,F_skn))

        del atm_tmp
        return float(F_skn - F_atm)
    
    log.info("Solving for global energy balance with conductive lid (T_magma = %.1f K)" % attrs["tmp_magma"])

    # Use an 'initial guess' method
    if method == 0:
        x0 = atm_inp.ts          # guess 1
        if T_surf_guess < minT:  # guess 2 (if not disabled)
            x1 = attrs["tmp_magma"] * 0.8
        else:
            x1 = T_surf_guess
        r = optimise.root_scalar(func, method='secant', x0=x0, x1=x1, xtol=atol, maxiter=20)

    # Use a 'bracketing' method
    elif method == 1:
        bracket = [800.0, maxT]
        if T_surf_max > minT:
            bracket = [bracket[0], min(T_surf_max, maxT)]
        r = optimise.root_scalar(func, method='brentq', bracket=bracket, xtol=atol, maxiter=20)

    else:
        raise Exception("Invalid solution method chosen (%d)" % method)

    # Extract solution
    T_surf_sol = float(r.root)
    succ  = bool(r.converged)

    if not succ:
        log.warning("Did not find solution for surface skin balance")
    else:
        log.info("Found surface solution")
        
    # Check bounds on T_surf
    T_surf = max(T_surf_sol, minT)
    T_surf = min(T_surf,     maxT)
    if T_surf_max > minT:
        T_surf = min(T_surf, T_surf_max)

    if T_surf != T_surf_sol:
        log.info("T_surf limits activated")
        log.info("    Found T_surf = %g K" % T_surf_sol)
        log.info("    Using T_surf = %g K" % T_surf)

    # Get atmosphere state from solution value
    atm = compute_moist_adiabat(ini_atm(T_surf), dirs, False, trppD, rscatter)
    atm.ts = T_surf

    if atm_bc == 0:
        F_atm = atm.net_flux[0]  
    else:
        F_atm = atm.net_flux[-1]  

    F_olr = atm.LW_flux_up[0]

    log.info("    T_surf = %g K"       % T_surf)
    log.info("    F_atm  = %.4e W m-2" % F_atm)
    log.info("    F_skn  = %.4e W m-2" % skin(atm))
    log.info("    F_olr  = %.4e W m-2" % F_olr)

    return atm




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 23 12:14:29 2023

@authors: 
Tim Lichtenberg (TL)    
Ryan Boukrouche (RB)
"""

import matplotlib.pyplot as plt
import numpy as np
import math 
#import json

from janus.modules.spectral_planck_surface import surf_Planck_nu
import janus.utils.GeneralAdiabat as ga # Moist adiabat with multiple condensibles

# Plotting
def plot_fluxes(atm,filename='output/fluxes.pdf'):

    fig,ax = plt.subplots()

    ax.axvline(0,color='black',lw=0.8)

    arr_p = atm.pl*1e-5

    ax.plot(atm.flux_up_total,arr_p,color='red',label='UP',lw=1)
    ax.plot(atm.SW_flux_up   ,arr_p,color='red',label='UP SW',linestyle='dotted',lw=2)
    ax.plot(atm.LW_flux_up   ,arr_p,color='red',label='UP LW',linestyle='dashed',lw=1)

    ax.plot(-1.0*atm.flux_down_total,arr_p,color='green',label='DN',lw=2)
    ax.plot(-1.0*atm.SW_flux_down   ,arr_p,color='green',label='DN SW',linestyle='dotted',lw=3)
    ax.plot(-1.0*atm.LW_flux_down   ,arr_p,color='green',label='DN LW',linestyle='dashed',lw=2)

    ax.plot(atm.net_flux ,arr_p,color='black',label='NET')

    ax.set_xlabel("Upward-directed flux [W m-2]")
    ax.set_xscale("symlog")
    ax.set_ylabel("Pressure [Pa]")
    ax.set_yscale("log")
    ax.set_ylim(top=np.amin(arr_p)/1.5, bottom=np.amax(arr_p)*1.5)

    ax.legend(loc='upper left')

    fig.savefig(filename, bbox_inches='tight', dpi=190)
    plt.close()




import numpy as np
import shutil , os
import subprocess
from scipy.interpolate import PchipInterpolator

import logging 
log = logging.getLogger("fwl."+__name__)


from .. import set_socrates_env 

def PrepareStellarSpectrum(wl, fl, star_file, nbins_max=95000):
    """Write a stellar spectrum.

    This function supplements InsertStellarSpectrum by writing a stellar 
    spectrum to a file in the format that SOCRATES expects. The flux needs to 
    be provided at 1 AU. Bins the spectrum down to at most nbins_max bins.

    Parameters
    ----------
        wl : list
            Wavelength [nm]
        fl : list
            Flux at 1 AU [erg s-1 cm-2 nm-1]
        star_file : str
            Path to output file, which will contain the stellar spectrum.
        nbins_max : int
            Number of spectral bins to down-sample to (-1 for no rebinning)
            
    """

    socrates_nbins_max = int(1e5 - 1)

    # Validate
    if (len(wl) != len(fl)):
        raise Exception("Stellar wavelength and flux arrays have different lengths")
    if (len(wl) < 500):
        log.warning("Loaded stellar spectrum is very short!")

    nbins_max = min(len(wl), nbins_max)
    
    if (nbins_max > socrates_nbins_max):
        raise Exception("Too many bins requested for stellar spectrum (maximum is %d)" % socrates_nbins_max)

    # Down-sample spectrum when necessary or requested
    if len(wl) > socrates_nbins_max:
        log.info("Rebinning stellar spectrum")

        # Store old wl,fl arrays
        wl_orig = wl
        fl_orig = fl

        if (nbins_max < 500):
            log.warning("Requested number of bins is small (%d bins)" % nbins_max)

        nbins_max = min( int(socrates_nbins_max), nbins_max) # Must be fewer than 100k

        # create interpolator on log-wavelength grid
        itp = PchipInterpolator(np.log10(wl_orig), fl_orig)
        wl = np.linspace(np.log10(wl_orig[0]), np.log10(wl_orig[-1]), nbins_max)

        # interpolate new fluxes
        fl = itp(wl)
        wl = 10**wl
            
    # Convert units
    wl = np.array(wl) * 1.0e-9  # [nm] -> [m]
    fl = np.array(fl) * 1.0e6   # [erg s-1 cm-2 nm-1] -> [W m-3]

    # Store header
    content = ""
    content += "Star spectrum at 1 AU. Created using PrepareStellarSpectrum() \n"
    content += "      WAVELENGTH        IRRADIANCE\n"
    content += "          (m)               (W/m3)\n"
    content += "*BEGIN_DATA\n"

    # Store body of data
    for i in range(len(wl)):
        content += str("      %1.7e      %1.7e\n" % (wl[i],fl[i]))

    # Store footer
    content += "*END\n"
    content += " "

    # Write content to file
    with open(star_file,'w') as handle:
        handle.write(content)


def InsertStellarSpectrum(orig_file:str, star_file:str, output_folder:str):
    """Insert a stellar spectrum.

    It's nice to be able to switch out the stellar spectrum for a different one. 
    This function takes in an original spectral file, with opacity data, and 
    inserts a stellar spectrum into a copy of it.

    Parameters
    ----------
        orig_file : str
            Path to original spectral file WITHOUT stellar spectrum.
        star_file : str
            Path to file containing stellar spectrum in the SOCRATES format.
        output_folder : str
            Path to output folder
            
    """

    # k files
    orig_filek = orig_file+"_k"

    outp_file  = os.path.join(output_folder, "star.sf")
    outp_filek = outp_file+"_k"

    # Delete "new" files if they already exist
    if os.path.exists(outp_file):
        os.remove(outp_file)
    if os.path.exists(outp_filek):
        os.remove(outp_filek)

    # Copy original files to new location (retain old files)
    shutil.copyfile(orig_file,  outp_file)
    shutil.copyfile(orig_filek, outp_filek)

    # Run prep_spec
    inputs = [outp_file,'a',                    # append existing file 
              '6','n','T','100 4000','250',     # tabulate thermal source function
              '2','n',star_file,'y',            # insert stellar spectrum
              '-1','EOF'                        # save and exit
              ]
    p = subprocess.run(['prep_spec'], stdout=subprocess.PIPE, input='\n'.join(inputs), encoding='ascii')
    if (p.returncode != 0):
        log.warning("prep_spec returned with code %d" % p.returncode)
    




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 23 12:01:27 2023

@authors: 
Tim Lichtenberg (TL)    
Ryan Boukrouche (RB)
"""

import copy
import numpy as np
from janus.modules.find_tropopause import find_tropopause
from janus.modules.set_stratosphere import set_stratosphere
from janus.modules.water_cloud import simple_cloud
from janus.modules.relative_humidity import compute_Rh

import logging
log = logging.getLogger("fwl."+__name__)

import janus.utils.GeneralAdiabat as ga # Moist adiabat with multiple condensibles
import janus.utils.socrates as socrates

def compute_moist_adiabat(atm, dirs, standalone, trppD, rscatter=False):
    """Compute moist adiabat case

    Parameters
    ----------
        atm : atmos
            Atmosphere object from atmosphere_column.py
        dirs : dict
            Named directories
        standalone : bool
            Running JANUS as standalone code?
        trppD : bool 
            Calculate tropopause dynamically?
        rscatter : bool
            Include Rayleigh scattering?
            
    """

    atm_moist = copy.deepcopy(atm)

    # Build general adiabat structure
    atm_moist = ga.general_adiabat(atm_moist)
    
    atm_moist.rh = compute_Rh(atm_moist)

    if atm.do_cloud:
        atm_moist = simple_cloud(atm_moist) # Before radiation, set up the cloud for Socrates using the current PT profile

    # Run SOCRATES
    atm_moist = socrates.radCompSoc(atm_moist, dirs, recalc=False, rscatter=rscatter)

    if standalone == True:
        log.info("w/o stratosphere (net, OLR): %.3f, %.3f W/m^2" % ( atm_moist.net_flux[0] , atm_moist.LW_flux_up[0]))

    # Calculate tropopause
    if (trppD == True) or (atm_moist.trppT > atm_moist.minT):
      
        # Find tropopause index
        atm_moist = find_tropopause(atm_moist,trppD, verbose=standalone)

        # Reset stratosphere temperature and abundance levels
        atm_moist = set_stratosphere(atm_moist)

        if atm.do_cloud:
            atm_moist = simple_cloud(atm_moist) # Update cloud location after previous PT changes
        # Recalculate fluxes w/ new atmosphere structure
        atm_moist = socrates.radCompSoc(atm_moist, dirs, recalc=True, rscatter=rscatter)

        if standalone == True:
            log.info("w/ stratosphere (net, OLR): %.3f, %.3f W/m^2" % (atm_moist.net_flux[0] , atm_moist.LW_flux_up[0] )) 

    return atm_moist



import pytest
from helpers import get_atmosphere_config, get_spectrum_data, work_directory
from janus.modules import MCPA_CBL
from numpy.testing import assert_allclose

TEST_DATA = (
    (0.3, (2.34297e03, 1.94397e03, -4.67185e01, 3.00023e03, 4.19568e02)),
#    (0.483333, (9.02641e02, 8.43916e02, 7.70272e01, 2.99961e03, 3.30552e02)),
#    (0.666667, (4.74451e02, 5.01731e02, 9.86425e01, 2.99951e03, 2.81454e02)),
#    (0.85, (2.91857e02, 4.21028e02, 1.73069e02, 2.99913e03, 2.49260e02)),
#    (1.03333, (1.97482e02, 4.16319e02, 2.48540e02, 2.99876e03, 2.26070e02)),
#    (1.21667, (1.42451e02, 4.15915e02, 2.94890e02, 2.99853e03, 2.08342e02)),
    (1.4, (1.07585e02, 4.15768e02, 3.24365e02, 2.99838e03, 1.94222e02)),
)


@pytest.mark.parametrize("inp,expected", TEST_DATA)
def test_instellation(tmpdir, inp, expected):
    out_drc = str(tmpdir) + "/"

    band_edges = get_spectrum_data(out_drc)
    atm, cfg = get_atmosphere_config(
        band_edges=band_edges,
        distance=inp,
        cfg_name="config_instellation.toml",
    )
    atm.setTropopauseTemperature()

    with work_directory(tmpdir):
        atm = MCPA_CBL(
            {"output": out_drc},
            atm,
            False,
            rscatter=True,
            T_surf_max=9.0e99,
            T_surf_guess=atm.trppT + 100,
        )

    ret = (atm.SW_flux_down[0], atm.LW_flux_up[0], atm.net_flux[0], atm.ts, atm.trppT)

    assert_allclose(ret, expected, rtol=1e-5, atol=0)



# atmosphere_column.py
# class for atmospheric column data

import toml
import numpy as np
import netCDF4 as nc
from janus.utils import phys
from janus.utils.height import AtmosphericHeight
import os, copy, platform, shutil
import pwd
from datetime import datetime

class atmos:
    
    def __init__(self, T_surf: float, P_surf: float, P_top: float, pl_radius: float, pl_mass: float, 
                 band_edges:list, vol_mixing: dict = {}, vol_partial: dict = {},
                 req_levels: int = 100, water_lookup: bool=False, alpha_cloud:float=0.0,
                 trppT: float = 290.0, minT: float = 1.0, maxT: float = 9000.0, do_cloud: bool=False,
                 re: float=0., lwm: float=0., clfr: float=0.,
                 albedo_s: float=0.0, albedo_pl: float=0.175, zenith_angle: float=54.74
                 ):
        
        """Atmosphere class    
    
        Stores compositional and thermodynamic information for the column.    
        Also stores planetary parameters.   

        One of either vol_mixing or vol_partial must be passed in. 
        If vol_partial is passed, then the value of P_surf is recalculated as the sum of partial pressures.

        Parameters
        ----------
            T_surf : float
                Surface temperature [K]
            P_surf : float
                Surface pressure [Pa]
            P_top : float
                Pressure at top of column [Pa]
            pl_radius : float
                Radius of rocky part of planet [m]
            pl_mass : float
                Mass of rocky part of planet [kg]
            band_edges : list
                List of band edges in nm, ascending

            vol_mixing : dict
                Dictionary of volatiles (keys) and mixing ratios (values)
            vol_partial: dict
                Dictionary of volatiles (keys) and partial pressures (values)

            req_levels : int
                Requested number of vertical levels
            alpha_cloud : float 
                Condensate retention fraction (1 -> Li et al 2018; 0 -> full rainout)
            water_lookup : bool
                Use lookup table for water thermodynamic values (e.g. L, c_p)
            trppT : float
                Tropopause temperature
            minT : float
                Temperature floor
            maxT : float
                Temperature ceiling
            re : float
                Effective radius of cloud droplets [m]
            lwm : float
                Liquid water mass fraction [kg/kg]
            clfr : float
                Water cloud fraction [adimensional]
            albedo_s : float
                Surface albedo
            albedo_pl : float
                Bond albedo (scattering) applied to self.toa_heating in socrates.py
            zenith_angle : float
                solar zenith angle, Hamano+15 (arccos(1/sqrt(3) = 54.74), Wordsworth+10: 48.19
        """

        # Parse volatiles 
        if (len(vol_mixing) == 0) and (len(vol_partial) == 0):
            raise Exception("Either vol_mixing OR vol_partial must be passed to atmos.__init__ function!\nNeither were.")
        if (len(vol_mixing) > 0) and (len(vol_partial) > 0):
            raise Exception("Either vol_mixing OR vol_partial must be passed to atmos.__init__ function!\nBoth were.")
        
        if len(vol_mixing) > 0:  # Set by mixing ratio
            self.ps = P_surf
            if (P_surf <= 0.0):
                raise Exception("Surface pressure passed to atmos.__init__ must be positive!\nValue passed = '%g'" % P_surf)
            
            tot_mixing =  float(sum(vol_mixing.values()))  # Ensure mixing ratios add up to unity
            self.vol_list = {}
            for key in vol_mixing.keys():
                self.vol_list[key] = vol_mixing[key]/tot_mixing

        
        if len(vol_partial) > 0: # Set by partial pressure
            self.ps = float(sum(vol_partial.values()))
            self.vol_list = {}
            for key in vol_partial.keys():
                self.vol_list[key] = vol_partial[key]/self.ps

        # Required volatiles
        required_vols = {"H2O","CO2","N2"}
        if len(required_vols.intersection(self.vol_list.keys())) < len(required_vols):
            raise Exception("Missing required volatiles!\nRequired vols = %s" % str(required_vols))
        
        # H2O floor to prevent NaNs
        self.vol_list["H2O"] = np.max( [ self.vol_list["H2O"], 1e-20 ] )
        
        # Remove volatiles with mixing ratio of zero
        # vol_list_old = copy.deepcopy(self.vol_list)
        # self.vol_list = {}
        # for key in vol_list_old.keys():
        #     if (vol_list_old[key] > 1.0e-30) or (key in required_vols):
        #         self.vol_list[key] = vol_list_old[key]


        # Initialise other variables
        self.alpha_cloud 	= alpha_cloud 	    	# The fraction of condensate retained in the column
        
        self.ts 			= T_surf		# Surface temperature, K

        self.ptop 			= P_top 			# Top pressure in Pa
        self.nlev 			= 10000  	   	# Number of vertical levels for adiabat integration
        self.step    		= 0.01  		# Adjust to match self.nlev
        self.nlev_save		= int(max(req_levels, 10))  		# Number of levels to save object
        self.p 				= np.zeros(self.nlev) 	   		# np.ones(self.nlev)
        self.pl 			= np.zeros(self.nlev+1)    		# np.ones(self.nlev+1)484854

        self.trppT          = trppT                 # Fixed value [K]
        self.trppidx		= 0 				 	# Tropopause: idx
        self.trppP 			= 0 				 	# Tropopause: prs
        self.minT           = minT                  # Minimum temperature allowed [K]
        self.maxT           = maxT                  # Maximum ^

        self.dt 			= 0.5 							# days

        self.instellation   = 0. 							# Instellation at planet's orbital separation, W/m^2
        self.toa_heating    = 0.                            # ASF

        self.inst_sf        = 3.0/8.0                       # Scale factor applied to instellation (see Cronin+14 for definitions)
        self.albedo_s   	= albedo_s                      # surface albedo
        self.albedo_pl   	= albedo_pl                     # Bond albedo (scattering) applied to self.toa_heating in socrates.py
        self.zenith_angle  	= zenith_angle                  # solar zenith angle, Hamano+15 (arccos(1/sqrt(3) = 54.74), Wordsworth+10: 48.19 (arccos(2/3)), see Cronin+14 for definitions

        self.planet_mass    = pl_mass
        self.planet_radius  = pl_radius
        self.grav_s 		= phys.G*self.planet_mass/(self.planet_radius**2) # m s-2
        
        self.tmp 			= np.zeros(self.nlev)      		# self.ts*np.ones(self.nlev)
        self.tmpl 			= np.zeros(self.nlev+1)
        self.Rcp    		= 2./7. 						# standard earth air
        self.n_species 		= 16
        self.mixing_ratios 	= np.zeros([self.n_species,self.nlev])

        self.water_lookup   = water_lookup
        
        self.tmp_magma      = 3000.0
        self.skin_d         = 0.01 # m
        self.skin_k         = 2.0  # W m-1 K-1
        
        # Level-dependent quantities
        self.p_vol 			= {} # Gas phase partial pressures
        self.pl_vol 		= {} # Gas phase partial pressures
        self.x_gas 			= {} # Gas phase molar concentration
        self.x_cond         = {} # Condensed phase molar concentration
        self.grav_z			= np.zeros(self.nlev) # Local gravity
        self.z 				= np.zeros(self.nlev)   # Atmospheric height (centres)
        self.zl 	    	= np.zeros(self.nlev+1) # Atmospheric height (edges)
        self.mu 			= np.zeros(self.nlev) # Mean molar mass in level
        self.xd 			= np.zeros(self.nlev) # Molar concentration of dry gas
        self.xv 			= np.zeros(self.nlev) # Molar concentration of moist gas
        self.xc 			= np.zeros(self.nlev) # Molar concentration of condensed phase
        self.rho            = np.zeros(self.nlev) # Density of atmosphere at a given level
        self.ifatm 			= np.zeros(self.nlev) # Defines nth level to which atmosphere is calculated
        self.cp      		= np.zeros(self.nlev) # Mean heat capacity
        self.height_error   = True # error when doing hydrostatic integration?

        # Define T and P arrays from surface up
        self.tmp[0]         = self.ts         		# K
        self.tmpl[0]         = self.ts         		# K
        self.p[0]            = self.ps         		# Pa
        self.pl[0]           = self.ps         		# Pa
        self.z[0]           = 0         			# m
        self.zl[0]
        self.grav_z[0]      = self.grav_s 			# m s-2


        # Instantiate object dicts and arrays
        for vol in self.vol_list.keys():
            # Instantiate as zero
            self.p_vol[vol]      = np.zeros(self.nlev)
            self.pl_vol[vol]     = np.zeros(self.nlev+1)
            self.x_gas[vol]      = np.zeros(self.nlev)
            self.x_cond[vol]     = np.zeros(self.nlev)
        
            # Surface partial pressures
            self.p_vol[vol][0]   = self.ps * self.vol_list[vol]


        # Spectral file
        self.has_contfunc = False
        self.band_edges = np.array(band_edges)  # units of [nm]
        self.bands_set = bool(len(band_edges) > 1)
        if self.bands_set:
            self.nbands 	    = np.size(self.band_edges)-1
            self.band_centres 	= (self.band_edges[1:] + self.band_edges[:-1]) / 2
            self.band_widths 	= np.abs(np.diff(self.band_edges))

            # Radiation heating and fluxes
            self.LW_flux_up 			= np.zeros(self.nlev_save)				# W/m^2
            self.LW_flux_down 			= np.zeros(self.nlev_save)				# W/m^2
            self.LW_flux_net			= np.zeros(self.nlev_save)				# W/m^2
            self.LW_spectral_flux_up 	= np.zeros([self.nbands,self.nlev_save])	# W/m^2/(band)
            self.LW_spectral_flux_down 	= np.zeros([self.nbands,self.nlev_save])	# W/m^2/(band)
            self.LW_heating				= np.zeros(self.nlev_save)				# K/day
            self.SW_flux_up 			= np.zeros(self.nlev_save)				# W/m^2
            self.SW_flux_down 			= np.zeros(self.nlev_save)				# W/m^2
            self.SW_flux_net			= np.zeros(self.nlev_save)				# W/m^2
            self.SW_spectral_flux_up 	= np.zeros([self.nbands,self.nlev_save])	# W/m^2/(band)
            self.SW_spectral_flux_down 	= np.zeros([self.nbands,self.nlev_save])	# W/m^2/(band)
            self.SW_heating				= np.zeros(self.nlev_save)				# K/day
            self.flux_up_total			= np.zeros(self.nlev_save)				# W/m^2
            self.flux_down_total		= np.zeros(self.nlev_save)				# W/m^2
            self.net_flux				= np.zeros(self.nlev_save)				# W/m^2
            self.net_spectral_flux	 	= np.zeros([self.nbands,self.nlev_save])	# W/m^2/(band)
            self.net_heating 			= np.zeros(self.nlev_save) 				# K/day from socrates
            self.heat                   = np.zeros(self.nlev_save)               # K/day from *
            self.cff					= np.zeros([self.nbands,self.nlev_save]) # W/m^2/(band). Compute by recompiling SOCRATES after setting the flags on lines 346 and 347 of src/aux/l_run_cdf.F90 to TRUE.
            self.LW_flux_up_i 			= np.zeros([self.nbands,self.nlev_save]) # W/m^2/(band)

        # Cloud flags (socrates/bin/rad_pcf.f90) and input arrays
        self.do_cloud = do_cloud
        if self.do_cloud :            
            self.cloud_scheme 			= 2	# -C 2 = ip_cloud_mix_max : maximum/random overlap in a mixed column 
            self.cloud_representation   = 1 # -K 1 = ip_cloud_homogen : ice and water mixed homogeneously 
            self.droplet_type           = 5 # -d 5 
            self.solver                 = 16 # -v 16 = i_solver : solver used for the two-stream calculations, chosen from those defined in solver_pcf.f90.
            # The below three variables should be zero until a cloud scheme forms clouds
            self.re                     = np.zeros(self.nlev_save) # Effective radius of the droplets [m]
            self.lwm                    = np.zeros(self.nlev_save) # Liquid water mass fraction [kg/kg]
            self.clfr                   = np.zeros(self.nlev_save) # Water cloud fraction
            # Floats defined in SocRadConv and used in the cloud scheme
            self.effective_radius       = re
            self.liquid_water_fraction  = lwm
            self.cloud_fraction         = clfr
        else:
            self.cloud_scheme 			= 5 # -C 5 = ip_cloud_off : clear sky. Other flags will be ignored. In principle we shouldn't need to define the variables below in this case.
            self.cloud_representation   = 1 
            self.droplet_type           = 5  
            self.solver                 = 13
            self.re                     = np.zeros(self.nlev_save)
            self.lwm                    = np.zeros(self.nlev_save)
            self.clfr                   = np.zeros(self.nlev_save) 
            self.effective_radius       = 0.0
            self.liquid_water_fraction  = 0.0
            self.cloud_fraction         = 0.0

        if self.alpha_cloud < 1.0e-20:
            self.effective_radius       = 0.0
            self.liquid_water_fraction  = 0.0
            self.cloud_fraction         = 0.0

    #New contructor based on toml file
    @classmethod
    def from_file(atm, file: str, band_edges: list, vol_mixing: dict = {}, vol_partial: dict = {}):

        with open(file, 'r') as f:
          cfg = toml.load(f)

        # Set parameters according to toml file if key present
        # Otherwise set default values
        T_surf     = cfg['atmos']['T_surf']     if "T_surf"     in cfg['atmos'] else 0.0
        P_surf     = cfg['atmos']['P_surf']     if "P_surf"     in cfg['atmos'] else 1.0e5
        P_top      = cfg['atmos']['P_top']      if "P_top"      in cfg['atmos'] else 1.0
        pl_radius  = cfg['planet']['pl_radius'] if "pl_radius"  in cfg['planet'] else 6.371e6
        pl_mass    = cfg['planet']['pl_mass']   if "pl_mass"    in cfg['planet'] else 5.972e24
        #optional arguments
        req_levels = cfg['atmos']['req_levels'] if "req_levels" in cfg['atmos'] else 100
        alpha      = cfg['atmos']['alpha']      if "alpha"      in cfg['atmos'] else 0.
        trppT      = cfg['atmos']['trppT']      if "trppT"      in cfg['atmos'] else 290.0
        do_cloud   = cfg['atmos']['do_cloud']   if "do_cloud"   in cfg['atmos'] else False
        re         = cfg['atmos']['re']         if "re"         in cfg['atmos'] else 0.
        lwm        = cfg['atmos']['lwm']        if "lwm"        in cfg['atmos'] else 0.
        clfr       = cfg['atmos']['clfr']       if "clfr"       in cfg['atmos'] else 0.
        albedo_pl  = cfg['atmos']['albedo_pl']  if "albedo_pl"  in cfg['atmos'] else 0.175
        albedo_s   = cfg['atmos']['albedo_s']   if "albedo_s"   in cfg['atmos'] else 0.
        zenith_angle = cfg['atmos']['zenith_angle'] if "zenith_angle" in cfg['atmos'] else 54.74

        return atm(T_surf, P_surf, P_top, pl_radius, pl_mass,
                   band_edges, vol_mixing, vol_partial,
                   #optional arguments
                   req_levels=req_levels,
                   alpha_cloud=alpha,
                   trppT=trppT,
                   do_cloud=do_cloud,
                   re=re,
                   lwm=lwm,
                   clfr=clfr,
                   albedo_pl=albedo_pl,
                   albedo_s=albedo_s,
                   zenith_angle=zenith_angle)

    def setSurfaceTemperature(self, Tsurf: float):

        self.ts = Tsurf
        self.tmp[0] = Tsurf
        self.tmpl[0] = Tsurf

    def setSurfacePressure(self, Psurf: float):

        self.ps = Psurf
        self.p[0] = Psurf
        self.pl[0] = Psurf

    def setPlanetProperties(self, pl_radius:float, pl_mass:float):

        self.planet_radius = pl_radius
        self.planet_mass = pl_mass
        self.grav_s = phys.G*self.planet_mass/(self.planet_radius**2) # m s-2
        self.grav_z[0] = self.grav_s

    def setVolatiles(self, vol_mixing: dict):

        tot_mixing =  float(sum(vol_mixing.values()))  # Ensure mixing ratios add up to unity
        self.vol_list = {key: val/tot_mixing for key, val in vol_mixing.items()}

        # H2O floor to prevent NaNs
        self.vol_list["H2O"] = np.max( [ self.vol_list["H2O"], 1e-20 ] )

        # Update volatile surface partial pressure
        for vol in self.vol_list:
            self.p_vol[vol][0] = self.ps * self.vol_list[vol]

    def setTropopauseTemperature(self):

        T_eqm = (self.instellation * self.inst_sf * (1.0 - self.albedo_pl) /phys.sigma)**(1.0/4.0)
        self.trppT = T_eqm * (0.5**0.25)  # radiative skin temperature

    def write_PT(self,filename: str="output/PT.tsv", punit:str = "Pa"):
        """Write PT profile to file, with descending pressure.

        Useful for when run outside of PROTEUS and for debugging.
        Designed with VULCAN compatibility in mind.

        Parameters
        ----------
            filename : string
                Output filename
            punit : string
                Pressure unit to use. Options: Pa, bar, dyne/cm2, atm.

        """

        p_scalefactor = 1.0
        match punit:
            case "Pa":
                p_scalefactor = 1.0
            case "bar":
                p_scalefactor = 1.0e-5
            case "dyne/cm2":
                p_scalefactor = 1.0e5
            case "atm":
                p_scalefactor = 1.01325e5
            case _:
                raise Exception("Unrecognised pressure unit '%s'"%punit)

        p_save = np.array(self.p) * p_scalefactor
        T_save = np.array(self.tmp)

        X = np.array([p_save,T_save]).T[::-1]

        header = '# (%s)\t(K)\nPressure\tTemp' % punit

        np.savetxt(filename,X,fmt='%1.5e',header=header,comments='',delimiter='\t')


    def write_ncdf(self, fpath:str):
        """Write atmosphere arrays to a netCDF file

        Parameters
        ----------
            fpath : string
                Output filename
        """ 

        # Get SOCRATES version string 
        verpath = os.path.join(os.environ.get("RAD_DIR"),"version")
        with open(verpath, 'r') as hdl:
            SOCRATES_version = hdl.read().replace("\n","")

        # ----------------------
        # Calculate gravity and height (in case it hasn't been done already)
        self.z = AtmosphericHeight(self, self.planet_mass, self.planet_radius)

        self.zl = np.zeros(len(self.z)+1)
        for i in range(1,len(self.z)):
            self.zl[i] = 0.5 * (self.z[i-1] + self.z[i])
        self.zl[0] = 2*self.z[0] - self.zl[1] # estimate TOA height

        # ----------------------
        # Prepare NetCDF
        if os.path.exists(fpath):
            os.remove(fpath)

        ds = nc.Dataset(fpath, 'w', format='NETCDF4')

        # Metadata
        try:
            # Try to get the login using os.getlogin()
            username = os.getlogin()
        except OSError:
            # If os.getlogin() fails, try an alternative method
            username = pwd.getpwuid(os.getuid()).pw_name
        ds.username         = str(username)
        ds.JANUS_version    = "0.1"
        ds.SOCRATES_version = str(SOCRATES_version)
        ds.platform         = str(platform.system())
        ds.date             = str(datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
        ds.description      = 'JANUS atmosphere data'
        ds.hostname         = str(platform.node())

        # ----------------------
        # Variable sizes
        nlev_c = len(self.p)
        nlev_l = nlev_c + 1

        gas_list = [str(gas) for gas in self.vol_list.keys()]
        ngases = len(gas_list)

        nchars = 16

        # Create dimensions
        ds.createDimension('nlev_c', nlev_c)            # Cell centres
        ds.createDimension('nlev_l', nlev_l)            # Cell edges
        ds.createDimension('ngases', ngases)            # Gases
        ds.createDimension('nchars', nchars)            # Length of string containing gas names
        ds.createDimension('nbands', self.nbands)       # Number of bands in the spectral file
        ds.createDimension('nchannels', self.nbands)

        # ----------------------
        # Scalar quantities  
        #    Create variables
        var_tstar   = ds.createVariable('tmp_surf',      'f4');  var_tstar.units = "K"     # Ground temperature
        var_trppP   = ds.createVariable('trppP',         'f4');  var_trppP.units = "Pa" 
        var_inst    = ds.createVariable("instellation",  'f4');  var_inst.units = "W m-2"  # Solar flux at TOA
        var_s0fact  = ds.createVariable("inst_factor",   'f4');                            # Scale factor applied to instellation
        var_znth    = ds.createVariable("zenith_angle",  'f4');  var_znth.units = "deg"    # Scale factor applied to instellation
        var_albbond = ds.createVariable("bond_albedo",   'f4');                            # Bond albedo used to scale-down instellation
        var_toah    = ds.createVariable("toa_heating",   'f4');  var_toah.units = "W m-2"  # TOA SW BC
        var_tmagma  = ds.createVariable("tmagma",        'f4');  var_tmagma.units = "K"    # Magma temperature
        var_tmin    = ds.createVariable("tfloor",        'f4');  var_tmin.units = "K"      # Minimum temperature
        var_tmax    = ds.createVariable("tceiling",      'f4');  var_tmax.units = "K"      # Maximum temperature
        var_plrad   = ds.createVariable("planet_radius", 'f4');  var_plrad.units = "m"     # Value taken for planet radius
        var_gsurf   = ds.createVariable("surf_gravity",  'f4');  var_gsurf.units = "m s-2" # Surface gravity
        var_albsurf = ds.createVariable("surf_albedo",   'f4');                            # Surface albedo
        var_sknd    = ds.createVariable("cond_skin_d"   ,'f4');  var_sknd.units = "m"      # Conductive skin thickness
        var_sknk    = ds.createVariable("cond_skin_k"   ,'f4');  var_sknk.units = "W m-1 K-1"    # Conductive skin thermal conductivity

        #     Store data
        var_tstar.assignValue(self.ts)
        var_inst.assignValue(self.instellation)  
        var_toah.assignValue(self.toa_heating)
        var_trppP.assignValue(self.trppP)
        var_znth.assignValue(self.zenith_angle)
        var_s0fact.assignValue(self.inst_sf)
        var_albbond.assignValue(self.albedo_pl)
        var_tmagma.assignValue(self.tmp_magma)
        var_tmin.assignValue(self.minT)
        var_tmax.assignValue(self.maxT)
        var_plrad.assignValue(self.planet_radius)
        var_gsurf.assignValue(self.grav_s)
        var_albsurf.assignValue(self.albedo_s)
        var_sknd.assignValue(self.skin_d)
        var_sknk.assignValue(self.skin_k)


        # ----------------------
        # Layer quantities  
        #    Create variables
        var_p     = ds.createVariable('p',       'f4', dimensions=('nlev_c'));           var_p.units = "Pa"
        var_pl    = ds.createVariable('pl',      'f4', dimensions=('nlev_l'));           var_pl.units = "Pa"
        var_tmp   = ds.createVariable('tmp',     'f4', dimensions=('nlev_c'));           var_tmp.units = "K"
        var_tmpl  = ds.createVariable('tmpl',    'f4', dimensions=('nlev_l'));           var_tmpl.units = "K"
        var_z     = ds.createVariable('z',       'f4', dimensions=('nlev_c'));           var_z.units = "m"
        var_zl    = ds.createVariable('zl',      'f4', dimensions=('nlev_l'));           var_zl.units = "m"
        var_grav  = ds.createVariable('gravity', 'f4', dimensions=('nlev_c'));           var_grav.units = "m s-2"
        var_cp    = ds.createVariable('cp',      'f4', dimensions=('nlev_c'));           var_cp.units = "J/(kg K)"

        var_mmw   = ds.createVariable('mmw',     'f4', dimensions=('nlev_c'));           var_mmw.units = "kg mol-1"
        var_gases = ds.createVariable('gases',   'S1', dimensions=('ngases', 'nchars'))  # Names of gases
        var_mr    = ds.createVariable('x_gas',   'f4', dimensions=('nlev_c', 'ngases'))  # Mixing ratios per level
        var_cmr   = ds.createVariable('x_cond',  'f4', dimensions=('nlev_c', 'ngases'))  # Condensate mixing ratios per level
        var_pvol  = ds.createVariable('p_vol',   'f4', dimensions=('nlev_c', 'ngases')); var_pvol.units = "Pa"  # Gas phase partial pressures
        var_plvol = ds.createVariable('pl_vol',  'f4', dimensions=('nlev_l', 'ngases')); var_pvol.units = "Pa"
        if self.bands_set:
            var_fdl   = ds.createVariable('fl_D_LW', 'f4', dimensions=('nlev_l'));           var_fdl.units = "W m-2"
            var_ful   = ds.createVariable('fl_U_LW', 'f4', dimensions=('nlev_l'));           var_ful.units = "W m-2"
            var_fnl   = ds.createVariable('fl_N_LW', 'f4', dimensions=('nlev_l'));           var_fnl.units = "W m-2"
            var_fds   = ds.createVariable('fl_D_SW', 'f4', dimensions=('nlev_l'));           var_fds.units = "W m-2"
            var_fus   = ds.createVariable('fl_U_SW', 'f4', dimensions=('nlev_l'));           var_fus.units = "W m-2"
            var_fns   = ds.createVariable('fl_N_SW', 'f4', dimensions=('nlev_l'));           var_fns.units = "W m-2"
            var_fd    = ds.createVariable('fl_D',    'f4', dimensions=('nlev_l'));           var_fd.units = "W m-2"
            var_fu    = ds.createVariable('fl_U',    'f4', dimensions=('nlev_l'));           var_fu.units = "W m-2"
            var_fn    = ds.createVariable('fl_N',    'f4', dimensions=('nlev_l'));           var_fn.units = "W m-2"
            var_hr    = ds.createVariable('hrate',   'f4', dimensions=('nlev_c'));           var_hr.units = "K day-1"
            var_bul   = ds.createVariable('ba_U_LW','f4',  dimensions=('nbands', 'nlev_l')); var_bul.units = "W m-2"
            var_bdl   = ds.createVariable('ba_D_LW','f4',  dimensions=('nbands', 'nlev_l')); var_bdl.units = "W m-2"
            var_bnl   = ds.createVariable('ba_N_LW','f4',  dimensions=('nbands', 'nlev_l')); var_bnl.units = "W m-2"
            var_bus   = ds.createVariable('ba_U_SW','f4',  dimensions=('nbands', 'nlev_l')); var_bus.units = "W m-2"
            var_bds   = ds.createVariable('ba_D_SW','f4',  dimensions=('nbands', 'nlev_l')); var_bds.units = "W m-2"
            var_bns   = ds.createVariable('ba_N_SW','f4',  dimensions=('nbands', 'nlev_l')); var_bns.units = "W m-2"
            var_bmin  = ds.createVariable('bandmin','f4',  dimensions=('nbands'));           var_bmin.units = "m"
            var_bmax  = ds.createVariable('bandmax','f4',  dimensions=('nbands'));           var_bmax.units = "m"

        if self.has_contfunc:
            var_cff   = ds.createVariable('contfunc',     'f4', dimensions=('nbands', 'nlev_c')); var_cff.units = "W m-2 m-1" ## units??

        var_re    = ds.createVariable('re',      'f4', dimensions=('nlev_c'));           var_re.units = "m"
        var_lwm   = ds.createVariable('lwm',     'f4', dimensions=('nlev_c'));           var_lwm.units = "kg kg-1"
        var_clfr  = ds.createVariable('clfr',    'f4', dimensions=('nlev_c'));           var_clfr.units = "none"

        #     Store data
        var_p[:]    = self.p[:]
        var_pl[:]   = self.pl[:]
        var_tmp[:]  = self.tmp[:]
        var_tmpl[:] = self.tmpl[:]
        var_z[:]    = self.z[:]
        var_zl[:]   = self.zl[:]
        var_mmw[:]  = self.mu[:]
        var_grav[:] = self.grav_z[:]
        var_cp[:]   = self.cp[:]

        var_gases[:] = np.array([ [c for c in gas.ljust(nchars)[:nchars]] for gas in gas_list ] , dtype='S1')
        var_mr[:]    = np.array([ [ self.x_gas[gas][i] for i in range(nlev_c-1,-1,-1) ] for gas in gas_list  ]).T
        var_cmr[:]   = np.array([ [ self.x_cond[gas][i] for i in range(nlev_c-1,-1,-1) ] for gas in gas_list  ]).T
        var_pvol[:]  = np.array([ [ self.p_vol[gas][i] for i in range(nlev_c-1,-1,-1) ] for gas in gas_list  ]).T
        var_plvol[:] = np.array([ [ self.pl_vol[gas][i] for i in range(nlev_l-1,-1,-1) ] for gas in gas_list  ]).T

        if  self.bands_set:
            var_fdl[:] = self.LW_flux_down[:]
            var_ful[:] = self.LW_flux_up[:]
            var_fnl[:] = self.LW_flux_net[:]

            var_fds[:] = self.SW_flux_down[:]
            var_fus[:] = self.SW_flux_up[:]
            var_fns[:] = self.SW_flux_net[:]

            var_fd[:] = self.flux_down_total[:]
            var_fu[:] = self.flux_up_total[:]
            var_fn[:] = self.net_flux[:]

            var_hr[:] = self.net_heating[:]

            var_bul[:,:]  = self.LW_spectral_flux_up[:,:]
            var_bdl[:,:]  = self.LW_spectral_flux_down[:,:]
            var_bnl[:,:]  = var_bul[:,:] - var_bdl[:,:]
            
            var_bus[:,:]  = self.SW_spectral_flux_up[:,:]
            var_bds[:,:]  = self.SW_spectral_flux_down[:,:]
            var_bns[:,:]  = var_bus[:,:] - var_bds[:,:]

            var_bmin[:]    = self.band_edges[:-1] * 1e-9
            var_bmax[:]    = self.band_edges[1:] * 1e-9

        if self.has_contfunc:
            var_cff[:,:]   = self.cff[:,:]

        var_re[:]   = self.re[:]
        var_lwm[:]  = self.lwm[:]
        var_clfr[:] = self.clfr[:]

        # ----------------------
        # Close
        ds.close()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul 14 13:44:49 2020

@author: grahamr
"""

import numpy as np
import scipy.integrate as spint
import os
#%%

# Function with necessary information for each species to calculate Rayleigh scattering
def species_info(species):
    info_dict = {}
    if species.lower() == 'co2':
        info_dict['Delta'] = 0.0805 #Vardavas and Carver 1984 for CO2
        info_dict['A'] = 43.9e-5 #Allen's Astrophysical Quantities (2002) Table 5.2
        info_dict['B'] = 6.4e-3 #Allen's
        info_dict['mu'] = 44e-3 #kg/mol
        
    elif species.lower() == 'n2':
        info_dict['Delta'] = 0.0305 #V&C 84 for N2
        info_dict['A'] = 29.06e-5 #Allen's Table 5.2
        info_dict['B'] = 7.7e-3 #Allen's  Table 5.2
        info_dict['mu'] = 28e-3 #kg/mol
    elif species.lower() == 'h2o':
        cross_section_hydrogen_1um=2.49e-6 #m^2/kg; mass-weighted cross section of hydrogen at 1 micron from Pierrehumbert 2010 Table 5.2
        info_dict['cross_section'] = 0.3743 * cross_section_hydrogen_1um #m^2/kg; mass weighted rayleigh cross section of water, from same table as h2
        info_dict['normalization_wavelength'] = 1e-6 #m; 1 micron
    elif species.lower() == 'pluriel_h2o':
        cross_section_cm2_molecule = 2.5e-27 #normalization cross section in cm^2 / molecule
        info_dict['cross_section'] =  cross_section_cm2_molecule * 10**-4 * 6.022e23 / 18e-3 #converting to m^2/kg
        info_dict['normalization_wavelength'] = 0.6e-6 #m
    elif species.lower() == 'pluriel_co2':
        cross_section_cm2_molecule = 1.24e-26 #normalization cross section in cm^2/molecule
        info_dict['cross_section'] = cross_section_cm2_molecule * 10**-4 * 6.022e23 / 44e-3 # converting to m^2/kg
        info_dict['normalization_wavelength'] = 0.532e-6 #m
    return info_dict

# Function for calculating raleigh scattering coefficient at a given wavelength 
# for a species w/ given index of refraction info (info which is obtained from species_info function)
def cross_section(wavelength, info_dict):#wavelength in m 
    if 'Delta' in info_dict:
        Delta = info_dict['Delta']
        A = info_dict['A']
        B = info_dict['B']
        mu = info_dict['mu']
        microns = wavelength * 10**6
        delta = (6+3*Delta)/(6-7*Delta) #V&C 1984
        coefficient = 4.577e-21 * delta/microns ** 4 * (A * (1 + B / microns**2)) ** 2 * (10**-4 * 6.022e23) / mu #V&C 1984; m^2/kg
    else:
        cross_section_1um = info_dict['cross_section']
        normalization_wavelength = info_dict['normalization_wavelength']
        coefficient = cross_section_1um * normalization_wavelength ** 4. / wavelength ** 4
    return coefficient


# This function integrates the Rayleigh cross-section for each species over a given band to allow for 
# proper averaging and then adds the integrated values together, weighted by molar mixing ratio
def band_integrator(species_list, molar_mixing_ratio_list, wavelength1_list, wavelength2_list):
    total_coefficient = 0
    for n, species in enumerate(species_list):
        info_tup = species_info(species)
        
        #Coefficient w/o mixing ratio weight
        partial_coefficient = np.vectorize(spint.quad,excluded=['args'])(cross_section,wavelength1_list,wavelength2_list, args = info_tup)[0]
        
        #Coefficient weighted  by mixing ratio
        total_coefficient +=  molar_mixing_ratio_list[n] * partial_coefficient
        
    return total_coefficient

# This is the function that adds  Rayleigh coefficients to spectral files
def rayleigh_coeff_adder(species_list = ['co2'], mixing_ratio_list = [1.], spectral_file_path= './spectral_files/sp_b318_HITRAN_a16_RS/sp_b318_HITRAN_a16',wavelength_dummy_file_path='./spectral_files/sp_b318_HITRAN_a16_RS/wavelength_band_file.txt'):
    spectral_file = open(spectral_file_path,'r')
    
    #Load a dummy file that will hold the wavelength bands
    
    wavelength_band_file = open(wavelength_dummy_file_path,'wt')
    
    #Make a loop that runs through the lines in the spectral file until it finds BLOCK 1,
    #at which point it begins to "pay attention" and read the relevant lines w/ the wavelength data in that file
    #into a new file that stores the wavelength bands'''
    pay_attention = False
    
    beginning_block_text_list = []
    
    for line in spectral_file:
        
        if pay_attention:
            #If the line says END and pay_attention is True, then we've reached the
            #end of block 1, so the loop can terminate because we don't need to read
            # the rest of the spectral file
            if 'END' in line:
                break
            # The first couple of lines in the block  have unnecessary text;
            # We ignore those by specifying that the loop only writes a line out
            # that starts with a space, which makes the loop ignore the unnecessary
            # bits and only grab the wavelength bands
            if line[0]==' ':
                wavelength_band_file.write(line)
            else:
                beginning_block_text_list.append(line)
        if '*BLOCK: TYPE =    1' in line:
            #The lines following this line will describe the wavelength data
            pay_attention = True
    
    spectral_file.close()
    wavelength_band_file.close()
    
    #Now we generate an array from the wavelength band file
    wavelength_bands = np.genfromtxt(wavelength_dummy_file_path,usecols=np.arange(0,3))
    #Delete the temporary wavelength band file
    os.remove(wavelength_dummy_file_path)
    #%%
    
    # This function calculates the rayleigh scattering cross section for a species at a given wavelength
    
    #calculate the change in wavelength over each band
    dlambda = wavelength_bands[:,2]-wavelength_bands[:,1]
    
    #Integrate the cross section function over the band
    cross_section_integral = band_integrator(species_list,mixing_ratio_list,wavelength_bands[:,1],wavelength_bands[:,2])
    
    #Divide each integral by dlambda to get the average cross section over each band
    cross_section_avgs = cross_section_integral/dlambda
    
    #Put the numerical labels and the cross section averages together in one array
    cross_section_list = np.vstack((wavelength_bands[:,0],cross_section_avgs)).T
    
    #%%
    #A list of the first couple of lines in BLOCK 3, which contains the Rayleigh scattering coefficients
    beginning_block_3 = ['*BLOCK: TYPE =    3: SUBTYPE =    0: VERSION =    0\n',\
                         'Rayleigh mass scatering coefficients at STP: unit m**2/kg\n',\
                         'Band        Rayleigh coefficient\n',\
                         '                 m2/kg\n']
    # This list will hold all the lines in Block 3
    block3_list = []
    # This loop adds the first few lines to block 3's list
    for n in range(len(beginning_block_3)):
        block3_list.append(beginning_block_3[n])
    
    # This loop puts the cross sections in the proper format and then adds them to the list line by line
    for n in range(cross_section_list.shape[0]):
        block3_list.append('{0:5d}'.format(np.array(cross_section_list[n,0],dtype=int))+'        %.9E\n'%(cross_section_list[n,1]))
    
    # Adding the end statement of block 3
    block3_list.append('*END\n')
    #%%
    # Opening the spectral file and reading all of its lines out
    spectral_file = open(spectral_file_path,'r')
    spectral_file_lines = spectral_file.readlines()
    spectral_file.close()
    
    # We're going to loop through the spectral file's lines and put them into lists
    # depending on whether they occur before or after block 3 -- 
    block3_flag = False
    block4_and_up_flag = False
    
    # Wwe make two more lists, the first will contain each line preceding block 3 and 
    # the second will contain each line after block 3
    first_blocks_list = []
    last_blocks_list = []
    for line in spectral_file_lines:
        
        #If we encounter this string in a line, we've entered block 3
        if '*BLOCK: TYPE =    3' in line:
            block3_flag = True
            
        # If we encounter this string in a line, we've entered block 4
        elif '*BLOCK: TYPE =    4' in line:
            block4_and_up_flag = True
            
            # block3_Flag is false now because we've left block 3
            block3_flag = False
            
        # Before we reach block 3 or block 4, lines go into first_blocks_list
        if not block3_flag and not block4_and_up_flag:
            first_blocks_list.append(line)
            
        # After we've made it through block 3 and entered block 4, lines go into
        # last_blocks_list
        elif block4_and_up_flag:
            last_blocks_list.append(line)
    
    #%%
    #now we create a dummy file to read all of the blocks into
    temp_spectral_file_path =  spectral_file_path+'_temp_spectral_file'
    temp_spectral_file = open(temp_spectral_file_path,'w')
    
    #First, blocks 0 to 2 are read in
    for n in range(len(first_blocks_list)):
        temp_spectral_file.write(first_blocks_list[n])
    
    #Then, block 3 is read in
    for n in range(len(block3_list)):
        temp_spectral_file.write(block3_list[n])
    
    # Finally, the remaining blocks are read in  
    for n in range(len(last_blocks_list)):
        temp_spectral_file.write(last_blocks_list[n])
        
    temp_spectral_file.close()
    #%%
    # The last step is to delete the original spectral file and replace it with the new one
    os.remove(spectral_file_path)
    os.rename(temp_spectral_file_path,spectral_file_path)

# For executing as standalone script
if __name__ == "__main__":

    rayleigh_coeff_adder(species_list = ['co2'], mixing_ratio_list = [1.], spectral_file_path= './spectral_files/sp_b318_HITRAN_a16_RS/sp_b318_HITRAN_a16',wavelength_dummy_file_path='./spectral_files/sp_b318_HITRAN_a16_RS/wavelength_band_file.txt')




# *****************************COPYRIGHT*******************************
# (C) Crown copyright Met Office. All rights reserved.
# For further details please refer to the file COPYRIGHT.txt
# which you should have received as part of this distribution.
# *****************************COPYRIGHT*******************************
'''
This file contains python functions that can be used to produce
multicolumn netCDF data compatible with l_run_cdf.
'''

from netCDF4 import Dataset
import numpy as np
import re

def write_dim(ncdf_file, n_val, vals, name, vtype, dims, units, title):
    dimension = ncdf_file.createDimension(name, n_val)
    variable = ncdf_file.createVariable(name, vtype, dims)
    variable.units = units
    variable.title = title
    variable[:] = vals

def write_var(ncdf_file, vals, name, vtype, dims, units, title):
    variable = ncdf_file.createVariable(name, vtype, dims)
    if (type(units) is str):
        variable.units = units
    if (type(title) is str):
        variable.title = title
    variable[:] = vals

def ncout_surf(file, lon, lat, basis, alb):

# ; Program to create netCDF files of surface albedo weights.
# ; Normally 'file' will have the extension .surf, and 'basis' = 1.
# ; 'alb' should then be an array of surface albedo values or a single value.

    levels = np.size(basis)
    n_lon = np.size(lon)
    n_lat = np.size(lat)
    nvals = np.size(alb)

    if (nvals == 1):
        albs = np.zeros(n_lon*n_lat*levels)
        albs += np.sum(alb)
    elif (nvals == n_lon*n_lat*levels):
        albs = alb
    else:
        raise RuntimeError(' Error in ncout_surf: arrays dont match', nvals
                            , n_lon * n_lat * levels)

    ncdf_file = create_cdf(file)

    write_dim(ncdf_file, n_lon, lon, 'lon', 'f4', 'lon', 'degree', 'LONGITUDE')
    write_dim(ncdf_file, n_lat, lat, 'lat', 'f4', 'lat', 'degree', 'LATITUDE')
    write_dim(ncdf_file, levels, basis, 'basis', 'i2', 'basis', 'None',
              'BASIS FUNCTION')

    write_var(ncdf_file, albs, 'alb', 'f4', ('basis', 'lat', 'lon'),'None',
              'ALBEDO WEIGHTS')

    ncdf_file.close()

#-------------------------------------------------------------------------------

def ncout_spectral_surf(file, lon, lat, bands, alb):

# Program to create netCDF files of surface albedo weights.
# Normally 'file' will have the extension .surf. 'bands' is the number bands.
# 'alb' should then be an array of surface albedo values for each band.

    basis = 1
    n_lon = np.size(lon)
    n_lat = np.size(lat)
    nvals = np.size(alb)

    if (nvals == 1):
        albs = np.zeros(n_lon * n_lat * np.sum(bands))
        albs += alb
    elif (nvals == n_lon * n_lat * bands):
        albs = alb
    else:
        raise RuntimeError(' Error in ncout_surf: arrays dont match', nvals
                           , n_lon * n_lat * bands)

    ncdf_file = create_cdf(file)
    write_dim(ncdf_file, n_lon, lon, 'lon', 'f4', 'lon', 'degree', 'LONGITUDE')
    write_dim(ncdf_file, n_lat, lat, 'lat', 'f4', 'lat', 'degree', 'LATITUDE')
    write_dim(ncdf_file, 1, basis, 'basis', 'i2', 'basis', 'None',
              'BASIS FUNCTION')
    write_dim(ncdf_file, np.sum(bands), np.arange(bands) + 1, 'bands', 'i2',
              'bands', 'None', 'BANDS')

    write_var(ncdf_file, albs, 'alb', 'f4', ('bands', 'basis', 'lat', 'lon'),
              'None', 'ALBEDO WEIGHTS')

    ncdf_file.close()

#-------------------------------------------------------------------------------

def ncout2d(file, lon, lat, val, name = None, longname = None
            , units = None):
# ; Program to create netCDF files of single level fields.
# ; For example (sol=1365.0 and lon, lat are arrays):
# ; ncout2d, 'out.stoa', lon, lat, sol, longname='Solar Irradiance', units='WM-2'
# ; (the optional argument 'name' is missing here and will be set from
# ; the file extension, i.e. name='stoa').

    n_lon = np.size(lon)
    n_lat = np.size(lat)
    nvals = np.size(val)
    if (nvals == 1):
        vals = np.zeros(n_lon*n_lat)
        vals += np.sum(val)
    elif (nvals == n_lon*n_lat):
        vals = val
    else:
        raise RuntimeError(' Error in ncout2d: arrays dont match', nvals
                           , n_lon * n_lat)

    if (type(name) is not str):
        patterns = re.compile(r'\\|\.')
        name = file[patterns.search(file).start() + 1:]

    ncdf_file = create_cdf(file)

    write_dim(ncdf_file, n_lon, lon, 'lon', 'f4', 'lon', 'degree', 'LONGITUDE')
    write_dim(ncdf_file, n_lat, lat, 'lat', 'f4', 'lat', 'degree', 'LATITUDE')

    write_var(ncdf_file, vals, name, 'f4', ('lat', 'lon'), units, longname)

    ncdf_file.close()

#-------------------------------------------------------------------------------

def ncout3d(file, lon, lat, p, val, name = None
            , longname = None, units = None):

# ; Program to create netCDF files of 3d fields on pressure levels.
# ; For example (lon, lat, p, and t are arrays):
# ; ncout3d, 'out.t', lon, lat, p, t, longname='Temperature', units='K'
# ; (the optional argument 'name' is missing here and will be set from
# ; the file extension, i.e. name='t').

    levels = np.size(p)
    n_lon = np.size(lon)
    n_lat = np.size(lat)
    nvals = np.size(val)

    if (nvals == 1):
        vals = np.zeros(n_lon * n_lat * levels).reshape(levels, n_lat, n_lon)
        vals += np.sum(val)
    elif (nvals == levels):
        vals = np.zeros(n_lon * n_lat * levels).reshape(levels, n_lat, n_lon)
        for i in np.arange(levels):
            vals[i, :, :] = val[i]
    elif (nvals == n_lon * n_lat * levels):
        vals = val.reshape(levels, n_lat, n_lon)
    else:
        raise RuntimeError(' Error in ncout3d: arrays dont match', nvals
                           , n_lon * n_lat * levels)

    if (p.size > 1):
        order = np.argsort(p)
        if (any (p[order] != p)):
            p = p[order]
            vals = vals[order,:,:]

    if (type(name) is not str):
        patterns = re.compile(r'\\|\.')
        name = file[patterns.search(file).start() + 1:]

    ncdf_file = create_cdf(file)

    write_dim(ncdf_file, n_lon, lon, 'lon', 'f4', 'lon', 'degree', 'LONGITUDE')
    write_dim(ncdf_file, n_lat, lat, 'lat', 'f4', 'lat', 'degree', 'LATITUDE')
    write_dim(ncdf_file, levels, p, 'plev', 'f4', 'plev', 'Pa', 'PRESSURE')

    title = None
    if (type(units) is not str):
        units = 'None'
    if (type(longname) is str):
        title = longname
    write_var(ncdf_file, vals, name, 'f4', ('plev', 'lat', 'lon'), units, title)

    ncdf_file.close()

#-------------------------------------------------------------------------------

def ncout_opt_prop(file, lon, lat, p, bands, absp, scat, phf):

# ; Program to create netCDF files of prescribed optical properties
# ; on pressure levels.
# ; For example (lon, lat, p, absp, scat and phf are arrays):
# ; ncout_opt_prop, 'out.op_soot', lon, lat, p, 6, absp, scat, phf

    levels = np.size(p)
    n_lon = np.size(lon)
    n_lat = np.size(lat)

    nvals = np.size(absp)
    if (nvals == levels*bands):
        absp.reshape(bands, levels)
        absp_vals = np.zeros(n_lon * n_lat * levels * bands).reshape(bands,
                    levels, n_lat, n_lon)
        for i in np.arange(n_lon):
            for j in np.arange(n_lat):
                absp_vals[:, :, j, i] = absp
    elif (nvals == n_lon*n_lat*levels*bands):
        absp_vals = absp.reshape(bands, levels, n_lat, n_lon)
    else:
        raise RuntimeError(' Error in ncout_opt_prop: absp arrays dont match'
                           , nvals, n_lon*n_lat*levels*bands)

    nvals = np.size(scat)
    if (nvals == levels*bands):
        scat.reshape(bands, levels)
        scat_vals = np.zeros(n_lon * n_lat * levels * bands).reshape(bands,
                    levels, n_lat, n_lon)
        for i in np.arange(n_lon):
            for j in np.arange(n_lat):
                scat_vals[:, :, j, i] = scat
    elif (nvals == n_lon * n_lat * levels * bands):
        scat_vals = absp.reshape(bands, levels, n_lat, n_lon)
    else:
        raise RuntimeError(' Error in ncout_opt_prop: scat arrays dont match'
                           , nvals, n_lon * n_lat * levels * bands)

    nvals = np.size(phf)
    if (nvals == 1):
        phf_vals = np.zeros(n_lon * n_lat * levels * bands).reshape(bands, 1,
                   levels, n_lat, n_lon)
        phf_vals += np.sum(phf)
    elif (nvals == levels*bands):
        phf.reshape(levels, bands)
        temp = np.zeros(n_lon * n_lat * levels * bands).reshape(bands, levels,
                        n_lat, n_lon)
        for i in np.arange(n_lon):
            for j in np.arange(n_lat):
                temp[:, :, j, i] = phf
    elif (nvals == n_lon * n_lat * levels * bands):
        phf_vals = absp.reshape(bands, 1, levels, n_lat, n_lon)
    else:
        raise RuntimeError(' Error in ncout_opt_prop: phf arrays dont match'
                           , nvals, n_lon * n_lat * levels * bands)

    order=np.argsort(p)
    if (any (p[order] != p)):
        p = p[order]
        absp_vals = absp_vals[:, order, :, :]
        scat_vals = scat_vals[:, order, :, :]
        phf_vals = phf_vals[:, :, order, :, :]

    ncdf_file = create_cdf(file)

    write_dim(ncdf_file, n_lon, lon, 'lon', 'f4', 'lon', 'degree', 'LONGITUDE')
    write_dim(ncdf_file, n_lat, lat, 'lat', 'f4', 'lat', 'degree', 'LATITUDE')
    write_dim(ncdf_file, levels, p, 'plev', 'f4', 'plev', 'Pa', 'PRESSURE')
    write_dim(ncdf_file, 1, 1, 'mom', 'i2', 'mom', 'none', 'moment')
    write_dim(ncdf_file, bands, np.arange(bands)+1, 'band', 'i2', 'band',
              'none', 'band')

    write_var(ncdf_file, absp_vals, 'abs', 'f4', ('band', 'plev', 'lat', 'lon')
              , 'M-1', 'absorption')
    write_var(ncdf_file, scat_vals, 'scat', 'f4', ('band', 'plev', 'lat', 'lon')
              , 'M-1', 'scattering')
    write_var(ncdf_file, phf_vals, 'phf', 'f4', ('band', 'mom', 'plev', 'lat',
              'lon'), 'none', 'phase function')

    ncdf_file.close()

#-------------------------------------------------------------------------------

def ncout_view(file, lon, lat, direction, level, pol, azim, rlev):

# Program to create netCDF ".view" files.

    n_lon = np.size(lon)
    n_lat = np.size(lat)
    n_dir  = np.size(direction)
    n_lvl  = np.size(level)
    n_pol  = np.size(pol)
    n_azm  = np.size(azim)
    n_rlv  = np.size(rlev)

    if (n_pol == 1):
        pols = np.zeros(n_lon * n_lat * n_dir).reshape(n_dir, n_lat, n_lon)
        pols += np.sum(pol)
    elif (n_pol == n_lon * n_lat * n_dir):
        pols = pol
    else:
        raise RuntimeError(' Error in ncout_view: arrays dont match'
                           , n_pol, n_lon * n_lat * n_dir)

    if (n_azm == 1):
        azims = np.zeros(n_lon*n_lat*n_dir).reshape(n_dir, n_lat, n_lon)
        azims += np.sum(azim)
    elif (n_azm == n_lon * n_lat * n_dir):
        azims = azim
    else:
        raise RuntimeError(' Error in ncout_view: arrays dont match'
                           , n_azm, n_lon * n_lat * n_dir)

    if (n_rlv == 1):
        rlevs = np.zeros(n_lvl)
        rlevs += np.sum(rlev)
    elif (n_rlv == n_lvl):
        rlevs = rlev
    else:
        raise RuntimeError(' Error in ncout_view: arrays dont match'
                           , n_rlv, n_lvl)

    ncdf_file = create_cdf(file)

    write_dim(ncdf_file, n_lon, lon, 'lon', 'f4', 'lon', 'degree', 'LONGITUDE')
    write_dim(ncdf_file, n_lat, lat, 'lat', 'f4', 'lat', 'degree', 'LATITUDE')
    write_dim(ncdf_file, levels, p, 'level', 'i2', 'level', 'None', 'LEVEL')
    write_dim(ncdf_file, n_dir, direction, 'direction', 'i2', 'direction', 'None', 'MOMENT')

    write_var(ncdf_file, pols, 'pol', 'f4', ('direction', 'lat', 'lon')
              , 'degree', 'POLAR VIEWING ANGLE')
    write_var(ncdf_file, azims, 'azim', 'f4', ('direction', 'lat', 'lon')
              , 'degree', 'AZIMUTHAL VIEWING ANGLE')
    write_var(ncdf_file, rlevs, 'rlev', 'f4', ('level')
              , 'None', 'VIEWING LEVEL')

    ncdf_file.close()

#-------------------------------------------------------------------------------

def create_cdf(path):
    ncdf_file = Dataset(path, 'w', format = 'NETCDF3_CLASSIC', clobber = 'true')
    return ncdf_file



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 23 11:57:03 2023

@authors: 
Tim Lichtenberg (TL)    
Ryan Boukrouche (RB)
"""

import numpy as np
import janus.utils.phys as phys
import janus.utils.GeneralAdiabat as ga # Moist adiabat with multiple condensibles
    
def set_stratosphere(atm):

    trpp_idx = int(atm.trppidx)
    trpp_prs = atm.trppP
    trpp_tmp = atm.trppT

    if trpp_idx < 0:
        return atm

    # Standard nodes
    for prs_idx, prs in enumerate(atm.p):
        if prs <= trpp_prs:
            atm.tmp[prs_idx] = trpp_tmp
    atm.tmp = np.clip(atm.tmp,atm.minT,None)

    # Staggered nodes
    for prsl_idx, prls in enumerate(atm.pl):
        if prls <= trpp_prs:
            atm.tmpl[prsl_idx] = trpp_tmp
    atm.tmpl = np.clip(atm.tmpl,atm.minT,None)

    # Set mixing ratios to same as tropopause
    for idx in reversed(range(0, trpp_idx)):
    
        atm.cp[idx] = 0.
        atm.mu[idx] = 0.

        # Volatile abundances
        for vol in atm.vol_list.keys():
        
            # Saturation vapor pressure
            p_vol_sat     = ga.p_sat(vol, atm.tmp[idx], water_lookup=atm.water_lookup)

            # If still condensible
            if atm.p[idx] > p_vol_sat:

                cond_diff            = (atm.x_cond[vol][idx] - atm.x_cond[vol][trpp_idx])

                atm.xc[idx]          -= cond_diff
                atm.xv[idx]          += cond_diff

                atm.x_cond[vol][idx] = atm.x_cond[vol][trpp_idx]
                atm.x_gas[vol][idx]  = atm.x_gas[vol][trpp_idx]
                atm.p_vol[vol][idx]  = atm.x_gas[vol][idx] * atm.p[idx]

            # If not anymore
            else:
                atm.xc[idx]          -= atm.x_cond[vol][idx]
                atm.x_gas[vol][idx]  = atm.x_gas[vol][trpp_idx]
                atm.xd[idx]          += atm.x_gas[vol][idx] + atm.x_cond[vol][idx]
                atm.xv[idx]          -= atm.x_gas[vol][idx]

                atm.x_cond[vol][idx] = 0.
                
                atm.p_vol[vol][idx]  = atm.x_gas[vol][trpp_idx] * atm.p[idx]
                

            # Renormalize cp w/ molar concentration
            atm.cp[idx]   += (atm.x_gas[vol][idx] + atm.x_cond[vol][idx]) * ga.cpv(vol, atm.tmp[idx]) / (atm.xd[idx] + atm.xv[idx] + atm.xc[idx]) # w/ cond

            # Renormalise mu with updated x_gas values 
            atm.mu[idx] += phys.molar_mass[vol] * atm.x_gas[vol][idx]

    return atm



from __future__ import annotations

import os
import subprocess as sp
import zipfile
from pathlib import Path

import click
import platformdirs
import requests

SOCRATES_DATA_DIR = Path(platformdirs.user_data_dir('socrates'))
SOCRATES_DIR = Path(os.environ.get('RAD_DIR', SOCRATES_DATA_DIR / 'SOCRATES'))


def _set_permissions(drc: Path):
    """Set executable flag for scripts."""
    # Set executable permissions for make script
    (drc / 'make' / 'mkdep').chmod(mode=33261)


def _download(*, url: str, filename: str):
    """Download file from url."""
    chunk_size = 1024 * 8

    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        total_size = int(r.headers.get('Content-Length', 0))
        with (
            open(filename, 'wb') as f,
            click.progressbar(label='Downloading', length=total_size) as pbar,
        ):
            for chunk in r.iter_content(chunk_size=chunk_size):
                f.write(chunk)
                pbar.update(chunk_size)


def download_socrates(ref: str = 'main'):
    """Version can be 'main' or the git commit hash."""
    filename = 'socrates.zip'
    path = 'refs/heads/main' if ref == 'main' else ref

    _download(
        url=f'https://github.com/nichollsh/SOCRATES/archive/{path}.zip',
        filename=filename,
    )

    SOCRATES_DATA_DIR.mkdir(exist_ok=True, parents=True)

    subdir = f'SOCRATES-{ref}'
    target_dir = SOCRATES_DATA_DIR / subdir

    click.echo(f'Extracting to {target_dir}')

    with zipfile.ZipFile(filename, 'r') as zip_ref:
        zip_ref.extractall(SOCRATES_DATA_DIR)

    _set_permissions(drc=target_dir)

    sp.run(['bash', 'configure'], cwd=target_dir)
    sp.run(['bash', 'build_code'], cwd=target_dir)

    symlink = SOCRATES_DATA_DIR / 'SOCRATES'
    symlink.unlink(missing_ok=True)
    symlink.symlink_to(target_dir)

    click.echo(f'SOCRATES downloaded to {target_dir}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 23 12:05:39 2023

@authors: 
Tim Lichtenberg (TL)    
Ryan Boukrouche (RB)
"""

import copy
import numpy as np

from janus.modules.dry_adiabat_setup import dry_adiabat_atm
from janus.modules.moist_adjustment_H2O import moist_adj
from janus.modules.dry_adjustment import DryAdj
from janus.modules.simple_boundary import simple_boundary_tend
from janus.modules.water_cloud import simple_cloud

import janus.utils.GeneralAdiabat as ga # Moist adiabat with multiple condensibles
import janus.utils.socrates as socrates
import janus.utils.phys as phys

# Time integration for n steps
def compute_dry_adiabat(atm, dirs, standalone, rscatter=False, pure_steam_adj=False, surf_dt=False, cp_surf=1e5, mix_coeff_atmos=1e6, mix_coeff_surf=1e6):

    # Dry adiabat settings 
    rad_steps   = 100  # Maximum number of radiation steps
    conv_steps  = 30   # Number of convective adjustment steps (per radiation step)
    dT_max      = 20.  # K, Maximum temperature change per radiation step
    T_floor     = 10.  # K, Temperature floor to prevent SOCRATES crash

    # Build general adiabat structure
    atm                 = ga.general_adiabat(copy.deepcopy(atm))

    # Copy moist pressure arrays for dry adiabat
    atm_dry             = dry_adiabat_atm(atm)

    # Initialise previous OLR and TOA heating to zero
    PrevOLR_dry         = 0.
    PrevMaxHeat_dry     = 0.
    PrevTemp_dry        = atm.tmp * 0.
    # Initialize the surface temperature tendency to 0 
    dT_surf             = 0

    # Time stepping
    for i in range(0, rad_steps):

        # atm.toa_heating = 0

        # Compute radiation, midpoint method time stepping
        try:
            if atm.do_cloud:
                atm_dry         = simple_cloud(atm_dry) # Before radiation, set up the cloud for Socrates using the current PT profile
                atm_dry         = socrates.radCompSoc(atm_dry, dirs, recalc=False, rscatter=rscatter)
                dT_dry          = atm_dry.net_heating * atm_dry.dt
    
                # Limit the temperature change per step
                dT_dry          = np.where(dT_dry > dT_max, dT_max, dT_dry)
                dT_dry          = np.where(dT_dry < -dT_max, -dT_max, dT_dry)
    
            # Do the surface balance
            if surf_dt:
                # Net surface flux (positive downward, negative upward)
                net_Fs = atm_dry.SW_flux_down[len(atm_dry.tmp)] + atm_dry.LW_flux_down[len(atm_dry.tmp)] - atm_dry.LW_flux_up[len(atm_dry.tmp)] - atm_dry.SW_flux_up[len(atm_dry.tmp)]
                dT_surf += net_Fs/cp_surf 
                dT_dry, dT_surf = simple_boundary_tend(len(dT_dry)-1,atm_dry.tmp,atm_dry.ts,dT_dry,dT_surf,mix_coeff_atmos,mix_coeff_surf)
                # Apply surface heating
                atm_dry.ts      += dT_surf * atm_dry.dt

                # Update temperature of lowest cell
                k_turb          = .1
                dT_dry[-1] += (atm_dry.ts - atm_dry.tmp[-1])/mix_coeff_atmos
                dT_surf    -= (atm_dry.ts - atm_dry.tmp[-1])/mix_coeff_surf
                atm_dry.tmp[-1] += dT_dry[-1] * k_turb * (atm_dry.tmp[-1] - atm_dry.ts) 
                
            # Apply heating
            atm_dry.tmp     += dT_dry 
            
            # Pure steam convective adjustment
            if pure_steam_adj:
                dT_moist = moist_adj(atm_dry,1000.)
                atm_dry.tmp     += dT_moist
            
            # Dry convective adjustment
            for iadj in range(conv_steps):
                atm_dry     = DryAdj(atm_dry)
    
            # Temperature floor to prevent SOCRATES crash
            if np.min(atm_dry.tmp) < T_floor:
                atm_dry.tmp = np.where(atm_dry.tmp < T_floor, T_floor, atm_dry.tmp)
    
            # Convergence criteria
            dTglobal_dry    = abs(round(np.max(atm_dry.tmp-PrevTemp_dry[:]), 4))
            dTtop_dry       = abs(round(atm_dry.tmp[0]-atm_dry.tmp[1], 4))
    
            # Break criteria
            dOLR_dry        = abs(round(atm_dry.LW_flux_up[0]-PrevOLR_dry, 6))
            dbreak_dry      = (0.01*(phys.sigma*atm_dry.ts**4)**0.5)
    
            # Inform during runtime
            if i % 2 == 1 and standalone == True:
                print("Dry adjustment step %d:" % (i+1))
                print("\tOLR = " + str(atm_dry.LW_flux_up[0]) + " W/m^2", "dOLR = " + str(dOLR_dry) + " W/m^2")
                print("\tT_top = " + str(atm_dry.tmp[0]) + " K,",  "dT_top = " + str(dTtop_dry) + " K")
                print("\tT_bot = " + str(atm_dry.tmp[-1]) + " K,",  "dT_max = " + str(dTglobal_dry) + " K")
                print("\tT_surf = " +str(atm_dry.ts) + " K" )
                
    
            # Reduce timestep if heating is not converging
            if dTglobal_dry < 0.05 or dTtop_dry > dT_max:
                atm_dry.dt  = atm_dry.dt*0.99
                # if standalone == True:
                #     print("Dry adiabat not converging -> dt_new =", round(atm_dry.dt,5), "days")
    
            # Sensitivity break condition
            if (dOLR_dry < dbreak_dry) and i > 5:
                if standalone == True:
                    print("Timestepping break ->", end=" ")
                    print("dOLR/step =", dOLR_dry, "W/m^2, dTglobal_dry =", dTglobal_dry)
                break    # break here
        except:
            if standalone == True:
                print("Socrates cannot be executed properly, T profile:", atm_dry.tmp)
            break    # break here

        PrevOLR_dry       = atm_dry.LW_flux_up[0]
        PrevMaxHeat_dry   = abs(np.max(atm_dry.net_heating))
        PrevTemp_dry[:]   = atm_dry.tmp[:]

    return atm_dry



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jan 12 17:18:48 2024

@authors:    
Ryan Boukrouche (RB)
"""

import numpy as np
from janus.utils.find_lcl import find_intersection
import janus.utils.GeneralAdiabat as ga
    
def compute_Rh(atm):
    """Compute the relative humidity.

    Parameters
    ----------
        atm : atmos
            Atmosphere object

    """

    #partial_p = atm.p_vol['H2O']-atm.x_cond['H2O']*atm.p # Remove the condensates so that only the vapor is accounted for.
    partial_p = atm.x_gas['H2O']*atm.p
    p_sat = [ga.p_sat('H2O', atm.tmp[i]) for i in range(len(atm.p))]
    R_h = partial_p/p_sat
    return R_h

def update_for_constant_RH(atm, R_h):
    """Compute the partial pressure and mixing ratio of the condensible that keeps the relative humidity constant below the LCL. Also update the total pressure.

    Parameters
    ----------
        atm : atmos
            Atmosphere object

        R_h : float
            Relative humidity

    """

    p_sat = [ga.p_sat('H2O', atm.tmp[i]) for i in range(len(atm.p))]

    i_lcl, lcl = find_intersection(p_sat, atm.p_vol['H2O'], tolerance=1e0)

    atm.p_vol['H2O'] = R_h * p_sat
    atm.p = sum(atm.p_vol[molecule] for molecule in atm.vol_list.keys()) # Updating the total pressure
    atm.ps = float(sum(atm.p_vol[molecule][-1] for molecule in atm.vol_list.keys())) # Not forgetting the surface pressure

    return atm



import os
import numpy as np

# Read band information from spectral file
def ReadBandEdges(spfile:str):

    if not os.path.exists(spfile):
        raise Exception("Canot find spectral file '%s'"%spfile)

    with open(spfile,'r') as hdl:
        lines = hdl.readlines()
    band_edgesm = []
    block_idx:int = -999999999
    for l in lines:
        # We want block 1 data
        if "Band        Lower limit         Upper limit" in l:
            block_idx = 0 
        if (block_idx > 0) and ("*END" in l):
            break 
        
        # Read bands
        if (block_idx > 0):
            s = [float(ss.strip()) for ss in l.split()[1:]]
            # First edge
            if block_idx == 1:
                band_edgesm.append(s[0])  # Note that these are in units of [metres]
            # Upper edges
            band_edgesm.append(s[1])  # [metres]

        # Block index 
        block_idx += 1
        
    return [v*1e9 for v in band_edgesm]  # convert to nm and return




#Edits and updates 6/2018:
#    *Replaced print with print function, imported from __future__
#    *Deprecated Ngl graphics
#    *Removed Numeric import
#    *Changed functions with variable positional arguments to use __code__
#     attribute instead of func_code
#Edits and updates 10/2018:
#    *Gave interp a __code__.co_argcount attribute so it can
#     be accepted by utilities that count arguments this way
#       (There is probably a better way to handle this. How to
#        count arguments for a callable object?)
#    *Removed all references to Numeric and Numeric.Float

#ToDo:  *Check for right column length in setitem and addCurve
#
#       *Implement show,hide for curves (in X() and Y())
#       *Implement missing data coding .
#            self.setMissingDataCode(code) (char or numeric)
#            possibly translate, check and force consistency
#       *Possibly handle missing data for plotting using a
#        fill() or interp() method to fill in missing data using
#        the interpolation routine. The best way to handle
#        missing data in computations is by using Masked Arrays
#
#       *Make and import a PathFinder module, which the instructor
#        will customize for the site.  This module will help the
#        students find locations of datasets and chapter scripts.
#        Could provide "links" to find script that produced a given
#        figure in the text.  Alternately, we could just
#        define directory strings like WorkbookDatasets here in
#        ClimateUtilities 
#
#

#For Python 2/3 compatibility
from __future__ import print_function

#-----------------------------------------------
#
#Import array package
#
#------------------------------------------------
import numpy

#-----------------------------------------------------
#Import graphics utilities
#
#First try to import ClimateGraphicsMPL. which contains the
#implementation of the plotting commands using MatPlotLib as a driver.
#use other graphics drivers. If MatPlotLib is not found a
#dummy graphics stub module is imported, which allows the courseware
#to be run without the user needing to explicitly comment out
#the plot calls in the Chapter Scripts.
#
#Formerly, PyNgl was supported as an alternate graphics driver,
#but MatPlotLib has become easy enough to install that PyNgl 
#support has been dropped, though for those using PyNgl it
#is not hard to adapt ClimateGraphicsMPL to use PyNgl instead
#------------------------------------------------------

from janus.utils.ClimateGraphicsMPL import * #Try importing MatPlotLib

#Section 1: -----Data handling utilities-------------------------------

# ToDo: Put documentation on use of Curve object here!
#
#       Add keyword arguments for axes, etc,
#
#       How to handle missing data codes? Note that
#         since lists are converted to numpy arrays,
#         text codes can't be used.
#       Provide a method to set which data is X.
#
#       Add output option to put out a LaTeX formatted table
#
#       Provide an easy way to add data column long-names
#
class Curve:
    def __init__(self):
         self.Xid = None # id of data to be considered as X
         self.data = {} # Dictionary of data columns
         self.label = {} #Data column label dictionary
         self.scatter = {} #Marker for making a curve a scatter plot
                           #i.e. suppress line drawing and plot symbol
         self.idList = [] #Keeps track of original order of data
         self.NumCurves = 0
         self.description = None # A string providing general information
         self.PlotTitle = '' # Title of the plot
         self.switchXY = 0 # Switches X and Y axes for plotting
         self.reverseX = 0 # Reverses X axis for plotting
         self.reverseY = 0 # Reverses Y axis for plotting
         self.XlogAxis = 0 # Use logarithmic axis for X
         self.YlogAxis = 0 # Use logarithmic axis for Y
         self.Xlabel = '' #X axis label
         self.Ylabel = '' #Y axis label
         #
         # Colors and line titles
         #
    #
    #Install a new curve in the data set, optionally with a variable name and label
    #Any 1D indexable object can be installed here:a numpy array, a Masked Array,
    #a Masked Variable (as in cdms) or an ordinary list. 
    def addCurve(self,data,id = '',label = ''):
        self.NumCurves += 1
        if len(id) == 0:
            id = 'v%d'%(self.NumCurves-1)
        #Transform data from list to a numpy array here
        if type(data) == type([]):
            data = numpy.array(data)
        self.data[id] = data
        if self.Xid == None:
            self.Xid = id   #Sets the default id for the X variable
        self.idList.append(id) #Keep track of order in which columns added
        self.label[id] = label
        self.scatter[id] = False
        #ToDo: Add checking for consistent lengths, type, etc. of what's being added
    def listVariables(self):
        return self.idList
    def __getitem__(self,id):
        return self.data[id]
    def __setitem__(self,id,data):
        try:
            n = len(data[:])
        except:
            print("Object on RHS is not indexable")
            return None
        #Transform data from list to a numpy array here
        if type(data) == type([]):
            data = numpy.array(data)
        if id in self.data.keys():
            self.data[id] = data
        else:
            self.addCurve(data,id)
    
    #Method to return numpy abcissa array for plotting.
    def X(self):
        #Use of cross section lets us get data from any indexed object
        #However, since Masked variables and Masked Arrays yield
        #their same types as cross sections, we have to check
        #explicitly for a _data component.
        #
        #This method is used mainly for generating arrays used in plotting,
        #and should be streamlined at some point. 
        temp = self.data[self.Xid][:]
        if hasattr(temp,'_data'):
            temp = temp._data[:]
        return numpy.array(temp,numpy.float)
    
    #Method to return numpy ordinate array for plotting
    def Y(self):
        #Use of cross section lets us get data from any indexed object
        outArray = []
        for id in self.idList:
            if not (id == self.Xid):
              column = self.data[id]
              if hasattr(column,'_data'):
                outArray.append(column._data[:]) #Deals with masked arrays and variables
              else:
                outArray.append(column[:])
        return numpy.array(outArray,numpy.float)

    #Dumps curve to a tab-delimited ascii file with column header
    def dump(self,fileName = 'out.txt'):
        outfile = open(fileName,'w')
        # Write out the data description if it is available.
        if not (self.description == None):
            if not self.description[-1] == '\n':
                self.description += '\n' #Put in a newline if needed
            outfile.write(self.description)
        header = ""
        fmt = ""
        ids = self.idList
        for id in ids:
            header += id+'\t'
            fmt += '%e\t'
        header = header[0:-1]+'\n'  # Replaces last tab with a newline.
        fmt = fmt[0:-1] + '\n'
        outfile.write(header)
        for i in range(len(self.data[ids[0]])):
            out = tuple([(self.data[id])[i] for id in ids])
            outfile.write(fmt%out)
        outfile.close()

    #Extracts a subset of the data and returns it as a new Curve.
    #This is useful if you only want to plot some of the columns.
    #The input argument dataList is a list of column names
    def extract(self,dataList):
        c = Curve()
        for dataName in dataList:
            c.addCurve(self[dataName],dataName)
        return c
            
            
            
            
# Scans a list of lines, locates data lines
# and size, splits of column headers and
# splits off general information text.
#
#A header can optionally be specified as input.
#The delimiter need not be specified if the file
#uses any whitespace character (including tabs) as
#column delimiters. Commas are not whitespace characters,
#and so need to be specified if they are used.
#
#ToDo:
#       *Implement missing data coding (with default '-').
#            self.setMissingDataCode(code) (char or numpy)
#            possibly translate, check and force consistency
#       *Replace optional positional arguments with keyword arguments
def scan(buff,inHeader=None,delimiter = None):
    if inHeader == None:
        inHeader = []
    #First delete blank lines
    buff = clean(buff)
    #
    #Now look for patterns that indicate data lines
    #     
    startDataLine,endDataLine = findData(buff)
    # Found number of items. Now read in the data
    #
    #Read in the first line. Is it a header?
    header = []
    if delimiter == None:
        line = buff[startDataLine].split()
    else:
        line = buff[startDataLine].split(delimiter)
    #
    try:
        slask = float(line[0])
    except:
        header = line
    if len(header) == 0:
        header = ['V%d'%i for i in range(len(line))]
        istart = 0
    else:
        istart = 1
    # Replace with inHeader if inHeader has been specified
    #  (Only use the input header if its length is consistent)
    if len(inHeader) == len(line):
        header = inHeader
    #
    # Read in the rest of the lines
    #
    varlist = [[] for i in range(len(header))]
    for line in buff[(startDataLine+istart):endDataLine]:
        if delimiter == None:
            items = line.split()
        else:
            items = line.split(delimiter)
        try:
         for i in range(len(varlist)):
            varlist[i].append(float(items[i]))
        except:
            print(items)
    vardict = {}
    for name in header:
        vardict[name] = varlist[header.index(name)]
    return vardict,header #header is returned so we can keep cols in orig order

#Eliminates blank lines
def clean(buff):
    buff = [line.strip() for line in buff]
    while(1):
      try:
         buff.remove('')
      except:
         break
    return buff

def findData(buff):
    runStarts = []
    for i in range(len(buff)-1):
        dn = abs(len(buff[i].split())-len(buff[i+1].split()))
        if not dn==0:
             runStarts.append(i+1)
    runStarts.append(len(buff))
    #Find index of run with max length
    nmax = -1
    for i in range(len(runStarts)-1):
        n = runStarts[i+1]-runStarts[i]
        if n > nmax:
            nmax = n
            imax = runStarts[i]
    #Deal with case where entire file is one run
    if nmax == -1:
        nmax = len(buff)
        imax = 0
    return imax,imax+nmax


#
# Function to read space or tab-delimited file into a curve object
# The input header is a list of names of the variables in each
# columns, which can be input optionally, mainly to deal with
# the case in which this information is not in the file being read
import string
def readTable(filename,inHeader = None,delimiter = None):
    f = open(filename)
    buff = f.readlines()
    data,header = scan(buff,inHeader,delimiter)
    c = Curve()
    for key in header:
        c.addCurve(data[key],key)
    return c

    
    
    
#==============================================
#---Section 2: Math utilities-------------------------------------------
#==============================================

#A dummy class useful for passing parameters.
#Just make an instance, then add new members
#at will.
class Dummy:
    pass

#---Polynomial interpolation and extrapolation (adapted
#from Numerical Recipes.
#
#It's used in Romberg extrapolation, but could be useful
#for polynomial OLR fits and so forth as well. Also
#needs online documentation
def polint(xa,ya,x):
    n = len(xa)
    if not (len(xa) == len(ya)):
            print("Input x and y arrays must be same length")
            return "Error"
        #Set up auxiliary arrays
    c = numpy.zeros(n,numpy.float)
    d = numpy.zeros(n,numpy.float)
    c[:] = ya[:]
    d[:] = ya[:]
    #Find closest table entry
    ns = 0
    diff = abs(xa[0]-x)
    for i in range(n):
            difft = abs(xa[i]-x)
            if difft < diff:
                diff = difft
                ns = i
    y=ya[ns]
    for m in range(1,n): 
        for i in range(n-m):
            ho=xa[i]-x
            hp=xa[i+m]-x
            w=c[i+1]-d[i]
            c[i] = ho*w/(ho-hp)
            d[i] = hp*w/(ho-hp)
        if 2*ns < (n-m):
            dy = c[ns]
        else:
            ns -= 1
            dy = d[ns]
        y += dy
        #You can also return dy as an error estimate. Here 
        #to keep things simple, we just return y.
    return y

#---------------------------------------------------------
#         Class for doing polynomial interpolation
#         from a table, using polint
#---------------------------------------------------------
#
#Usage:
#         Let xa be a list of independent variable
#         values and ya be a list of the corresponding
#         dependent variable values. Then, to create a function
#         f (actually a callable object, techically) that interpolates
#         or extrapolates to any value x, create f using
#                     f = interp(xa,ya)
#         Then you can get the value you want by invoking f(x)
#         for your desired x.
#
#         By default, the interpolator does fourth-order interpolation
#         using the four nearest neighbors. You can change this by
#         using an optional third argument to the creator. For
#         example
#
#                     f = interp(xa,ya,8)
#         will use the 8 nearest neighbors (if they are available)
#------------------------------------------------------------
class interp:
    def __init__(self,xa,ya,n=4):
        self.xa = numpy.array(xa)
        self.ya = numpy.array(ya)
        self.n = n
        self.__code__ = Dummy()
        self.__code__.co_argcount = 1#Gives it a function-like argcount
    def __call__(self,x):
        #Find the closes index to x
        if self.xa[0] < self.xa[-1]:
            i = numpy.searchsorted(self.xa,x)
        else:
            i = numpy.searchsorted(-self.xa,-x)
        i1 = max(i-self.n,0)
        i2 = min(i+self.n,len(self.xa))
        return polint(self.xa[i1:i2],self.ya[i1:i2],x)


#----Quadrature (definite integral) by Romberg extrapolation.
#**ToDo: Add documentation and help string

#Before developing a general quadrature class, we'll
#implement a class which efficiently carries out trapezoidal rule
#integration with iterative refinement
class BetterTrap:
    def __init__(self,f,params,interval,nstart):
        self.f = f
        self.n = nstart
        self.interval = interval
        self.params = params
        self.integral = self.dumbTrap(nstart)
    def dumbTrap(self,n):
        a = self.interval[0]
        b = self.interval[1]
        dx = (b-a)/n
        sum = dx*(self.f(a,self.params)+self.f(b,self.params))/2.
        for i in range(1,n):
            x = a+i*dx
            sum = sum + self.f(x,self.params)*dx
        return sum
    def refine(self):
        #Compute the sum of f(x) at the
        #midpoints between the existing intervals.
        #To get the refinement of the trapezoidal
        #rule sum we just add this to half the
        #previous result
        sum = 0.
        a = self.interval[0]
        b = self.interval[1]
        dx = (b-a)/self.n
        #Remember: n is the number of subintervals,
        #not the number of endpoints. Therefore we
        #have one midpoint per subinterval. Keeping that
        #in mind helps us get the range of i right in
        #the following loop
        for i in range(self.n):
            sum = sum + self.f(a+(i+.5)*dx,self.params)*(dx/2.)
        #The old trapezoidal sum was multiplied by
        #the old dx.  To get its correct contribution
        #to the refined sum, we must multiply it by .5,
        #because the new dx is half the old dx
        self.integral = .5*self.integral + sum
        #
        #Update the number of intervals
        self.n = 2*self.n


#Here I define a class called
#romberg, which assists in carrying out evaluation of
#integrals using romberg extrapolation. It assumes polint has
#been imported

class romberg:
    def __init__(self,f,nstart=4):
        self.nstart = nstart
        self.trap = None
        #
        #-------------------------------------------------
        #This snippit of code allows the user to leave the
        #parameter argument out of the definition of f
        #if it isn't needed
        #
        self.fin = f
        #Find the number of arguments of f and append a
        #parameter argument if there isn't any.
        nargs = f.__code__.co_argcount
        if nargs == 2:
            self.f = f
        elif nargs ==1:
            def f1(x,param):
                return self.fin(x)
            self.f = f1
        else:
            name = f.func_name
            print('Error: %s has wrong number of arguments'%name)
        #-----------------------------------------------------
        #
        #We keep lists of all our results, for doing
        #Romberg extrapolation. These are re-initialized
        #after each call
        self.nList = []
        self.integralList = []
    def refine(self):
        self.trap.refine()
        self.integralList.append(self.trap.integral)
        self.nList.append(self.trap.n)
        dx = [1./(n*n) for n in self.nList]
        return polint(dx,self.integralList,0.)
    #
    #Use a __call__ method to return the result. The
    #__call__ method takes the interval of integration
    #as its mandatory first argument,takes an optional
    #parameter argument as its second argument, and
    #an optional keyword argument specifying the accuracy
    #desired.
    #**ToDo: Introduce trick to allow parameter argument of
    #integrand to be optional, as in Integrator.  Also, make
    #tolerance into a keyword argument
    #
    def __call__(self,interval,params=None,tolerance=1.e-6):
        self.nList = []
        self.integralList = []
        #Make a trapezoidal rule integrator
        self.trap = BetterTrap(self.f,params,interval,self.nstart)
        self.nList.append(self.nstart)
        self.integralList.append(self.trap.integral)
        #
        #Refine initial evaluation until 
        oldval = self.refine()
        newval = self.refine()
        while abs(oldval-newval)>tolerance:
            oldval,newval = newval,self.refine()
        return newval
        
        

#-------Runge-Kutta ODE integrator
#**ToDo:
#      * Implement a reset() method which resets to initial conditions.
#         Useful for doing problem over multiple times with different
#         parameters.
#
#      * Referring to the independent variable as 'x' is awful, and
#        confusing in many contexts.  Introduce a variable name
#        dictonary with default names like 'independent' and 'dependent'
#        (and short synonyms) so if fi is the integrator object
#        fi['indep'] is the current value of the independent variable
#        and fi['dep'] is the current (vector) value of the dependent
#        variable. Then allow user to rename or synonym these to
#        the actual user-supplied names.  This is an alternative
#        to using the list returned by fi.next(). Then expunge
#        all references to things like fi.x from the examples and
#        chapter scripts. They are too confusing.  Similarly,
#        change the name of the increment from dx to something else
#
#      * Similarly, we could introduce a dictionary of some sort
#        to make it easier to set up multidimensional systems and
#        refer to the different vector components by name
#        (e.g. refer to v[0] at T, v[1] as dTdy , etc. )
#
#      * Make the integrator object callable. The call can return
#        a list of results for all the intermediate steps, or optionally
#        just the final value.  
class integrator:
  '''
  Runge-Kutta integrator, for 1D or multidimensional problems
  
  Usage:
  
  First you define a function that returns the
  derivative(s), given the independent and dependent
  variables as arguments. The independent variable (think
  of time) is always a scalar. The dependent variable (think
  of the x and y coordinates of a planet) can be either a scalar
  or a 1D array, according to the problem. For the
  multidimensional case, this integrator will work with any
  kind of array that behaves like a numpy array, i.e. supports
  arithmetic operations. It will not work with plain Python lists.
  The derivative function should return an array of derivatives, in
  the multidimensional case. The derivative function can have any name.
  
  The derivative function can optionally have a third argument, to provide
  for passing parameters (e.g. the radius of the Earth) to the
  function.  The "parameter" argument, if present, can be any Python
  entity whatsoever. If you need to pass multiple constants, or
  even tables or functions, to your derivative function, you can
  stuff them all into a list or a Python object.
  
  
  Example:
  In the 1D case, to solve the equation
                  dz/dt = -a*t*t/(1.+z*z)
  in which z is the dependent variable and t is the
  independent variable, your derivative function would  be
    def g(t,z):
       return -a*t*t/(1.+z*z)
  
  treating the parameter a as a global, or perhaps
    def g(t,z,params):
       return -params.a*t*t/(params.b+z*z)
  
  while in a 2D case, your function might look like:
    def g(t,z):
      return numpy.array([-z[1],z[0]])
  
  or perhaps something like:
    def g(t,z):
      return t*numpy.sin(z)
  
  or even
    def g(t,z,params):
      return numpy.matrixmultiply(params(t),z)
      
  where params(t) in this case is a function returning
  a numpy square matrix of the right dimension to multiply z.
  
  BIG WARNING:  Note that all the examples which return a
  numpy array return a NEW INSTANCE (i.e. copy) of an
  array.  If you try to set up a global array and re-use
  it to return your results from g, you will really be
  just returning a REFERENCE to the same array each time,
  and each call will change the value of all the previous
  results. This will mess up the computation of intermediate
  results in the Runge-Kutta step. An example of the sort of thing
  that will NOT work is:
    zprime = numpy.zeros(2,numpy.float)
    def g(t,z,params):
      zprime[0] = z[1]
      zprime[1] = -z[0]
      return zprime
  Try it out. This defines the harmonic oscillator, and a plot
  of the orbit should give a circle. However, it doesn't  The problem
  reference/value distinction.  The right way to define the function
  would be
    def g(t,z):
      return numpy.array([z[1],-z[0]])
  Try this one. It should work properly now. Note that any arithmetic
  performed on numpy array objects returns a new instance of an Array
  object. Hence, a function definition like 
    def g(t,z):
      return t*z*z+1.
  will work fine.
  
  Once you have defined the derivitave function, 
  you then proceed as follows.
  
  First c reate an integrator instance:
    int_g = integrator(g,0.,start,.01)
  
  where "0." in the argument list is the initial value
  for the independent variable, "start" is the initial
  value for the dependent variable, and ".01" is the
  step size. You then use the integrator as follows:
  
    int_g.setParams(myParams)
    while int_g.x < 500:
       print(int_g.next())
  
  The call to setParams is optional. Just use it if your
  function makes use of a parameter object. The next() method
  accepts the integration increment (e.g. dx) as an optional
  argument. This is in case you want to change the step size,
  which you can do at any time.  The integrator continues
  using the most recent step size it knows.
  
  Each call to int_g.next returns a list, the first of whose
  elements is the new value of the independent variable, and
  the second of whose elements is a scalar or array giving
  the value of the dependent variable(s) at the incremented
  independent variable. 
  
  '''
  def __init__(self, derivs,xstart,ystart,dx=None):
    self.derivsin = derivs
    #
    #The following block checks to see if the derivs
    #function has a parameter argument specified, and
    #writes a new function with a dummy parameter argument
    #appended if necessary. This allows the user to leave
    #out the parameter argument from the function definition,
    #if it isn't needed.
    nargs = derivs.__code__.co_argcount
    if nargs == 3:
        self.derivs = derivs
    elif nargs == 2:
        def derivs1(x,y,param):
            return self.derivsin(x,y)
        self.derivs = derivs1
    else:
        name = derivs.func_name
        print('Error: %s has wrong number of arguments'%name)
    #
    #
    self.x = xstart
    #The next statement is a cheap trick to initialize
    #y with a copy of ystart, which works whether y is
    #a regular scalar or a numpy array.  
    self.y = 0.+ ystart
    self.dx = dx #Can instead be set with the first call to next()
    self.params = None
  #Sets the parameters for the integrator (optional).
  #The argument can be any Python entity at all. It is
  #up to the user to make sure the derivative function can
  #make use of it.
  def setParams(self,params):
      self.params = params
  #Computes next step.  Optionally, takes the increment
  #in the independent variable as an argument.  The
  #increment can be changed at any time, and the most
  #recently used value is remembered, as a default
  def next(self,dx = None):
     if not (dx == None):
         self.dx = dx
     h = self.dx
     hh=h*0.5;
     h6=h/6.0;
     xh=self.x+hh;
     dydx = self.derivs(self.x,self.y,self.params)
     yt = self.y+hh*dydx
     dyt = self.derivs(xh,yt,self.params)
     yt =self.y+hh*dyt
     dym = self.derivs(xh,yt,self.params)
     yt =self.y+h*dym
     dym += dyt
     dyt = self.derivs(self.x+h,yt,self.params)
     self.y += h6*(dydx+dyt+2.0*dym)
     self.x += h
     return self.x,self.y



#**ToDo:  
#
#         Store the previous solution for use as the next guess(?)
#
#        Handle arithmetic exceptions in the iteration loop
#  
class newtSolve:
    '''
    Newton method solver for function of 1 variable
    A class implementing Newton's method for solving f(x) = 0.

    Usage: solver = newtSolve(f), where f is a function with
    calling sequence f(x,params). Values of x such that
    f(x,params) = 0 are
    then found by invoking solver(guess), where guess
    is the initial guess.  The solver returns the string
    'No Convergence' if convergence fails. The argument
    params allows parameters to be passed to the function.
    It can be left out of the function definition if you don't
    need it. Note that params can be any Python object at all
    (including,e.g.,lists, functions or more complex user-defined objects)

    Optionally, one can specify the derivative function
    in the creator,e.g. solver = newtSolve(f,fp).
    If the derivative function isn't specified, the solver
    computes the derivative approximately using a centered
    difference. Note that in either case you can access
    the derivative function by invoking solver.deriv(x)
    As for f, fp can be optionally defined with a parameter
    argument if you need it. The same parameter object is
    passed to f and fp. 

    Use solver.setParams(value) to set the parameter object
    Alternately, the parameter argument can be passed as
    an optional second argument in the solver call. (see
    example below).

    Adjustable constants:
     eps         Increment for computing numerical approximation to
                 the derivative of f
     tolerance   Accuracy criterion for ending the iteration
                 (an approximation to the error in the root)
     nmax        maximum number of iterations

    e.g. to change the maximum number of iterations for an instance
    of the class, set solver.nmax = 10 .
    ----------------Usage Examples-----------------------------

         Example 1: Function without parameters:
          def g(x):
              return x*x - 1.
          roots = newtSolve(g)
          roots(2.)
         
         Example 2, Function with parameters:
          def g(x,constants):
              return constants.a*x*x - constants.b
          roots = newtSolve(g)
          constants = Dummy()
          constants.a = 1.
          constants.b = 2.
          roots.setParam(constants)
          roots(2.)
          roots(1.)

         Example 2a:
         Instead of using roots.setParam(...) we could do
           roots(2.,constants)
           roots(1.)    the parameters are remembered
           constants.a = 3.
           roots(1.,constants)   We changed the constants

         Example 3, using scan to find initial guesses:
          def g(x):
              return x*x - 1.
          roots = newtSolve(g)
          guesses = roots.scan([-2.,2.],100)
          for guess in guesses:
              print roots(guess)
    '''
    def __init__(self,f,fprime = None):
        self.fin = f
        #Find the number of arguments of f and append a
        #parameter argument if there isn't any.
        nargs = f.__code__ .co_argcount
        if nargs == 2:
            self.f = f
        elif nargs ==1:
            def f1(x,param):
                return self.fin(x)
            self.f = f1
        else:
            name = f.func_name
            print('Error: %s has wrong number of arguments'%name)
        self.eps = 1.e-6
        def deriv(x,params):
            return (self.f(x+self.eps,params)- self.f(x-self.eps,params))/(2.*self.eps)
        if fprime == None:
            self.deriv = deriv 
        else:
            #A derivative function was explicitly specified
            #Check if it has a parameter argument
            nargs = fprime.__code__.co_argcount
            if nargs == 2:
                self.deriv = fprime #Has a parameter argument
            elif nargs == 1:
                self.fprimein = fprime
                def fprime1(x,param):
                    return self.fprimein(x)
                self.deriv = fprime1
            else:
                name = fprime.func_name
                print('Error: %s has wrong number of arguments'%name)
        self.tolerance = 1.e-6
        self.nmax = 100
        self.params = None
    def __call__(self,xGuess,params = None):
        if not (params == None):
            self.setParams(params)
        x = xGuess
        for i in range(self.nmax):
            dx = (self.f(x,self.params)/self.deriv(x,self.params))
            x = x - dx
            if abs(dx) < self.tolerance:
                return x
        return 'No Convergence'
    def setParams(self,params):
        #**ToDo: Check if f1 has a parameter argument
        #defined, and complain if it doesn't
        self.params = params
    def scan(self,interval,n=10):
        #Finds initial guesses to roots in a specified
        #interval, subdivided into n subintervals.
        #e.g. if the instance is called "solver"
        #solver.scan([0.,1.],100) generates a list
        #of guesses between 0. and 1., with a resolution
        #of .01. The larger n is, the less is the chance that
        #a root will be missed, but the longer the search
        #will take.  If n isn't specified, the default value is 10
        #
        #ToDo: Replace this with a bisection search, allowing user
        #to specify the maximum number of distinct guesses that
        #need to be found.
        guessList = []
        dx = (interval[1]-interval[0])/(n-1)
        flast = self.f(interval[0],self.params)
        for x in [interval[0]+ i*dx for i in range(1,n)]:
            fnow = self.f(x,self.params)
            if ((fnow >= 0.)&(flast <=0.)) or ((fnow <= 0.)&(flast >=0.)):
                guessList.append(x)
            flast = fnow
        return guessList





    
         
